/*->c.curveb */

#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"


#include "err.h"
#include "xext.h"
#include "poll.h"
#include "temp.h"
#include "alloc.h"
#include "fsx.h"
#include "trans.h"
#include "xmath.h"
#include "colour.h"
#include "deb.h"
#include "dbhi.h"
#include "bits.h"

#include "drawcurves.h"


#include "etc.h"



#include "constants.h"
#include "str.h"
#include "mhi.h"

#include "curvelo.h"


#ifdef NEVER


/*

typedef struct bezier
{
 coordstr left;
 coordstr leftc;
 coordstr rightc;
 coordstr right;
} bezier;





void bezdivide(bezier * p,bezier * a,bezier * b);


typedef os_error * (*bezflatfn)(int x0,int y0,int x1,int y1,void * data);


void bezflatten(bezier * p,int flatness,bezflatfn fn,void * data);


typedef os_error * (*bezlinefn)(int x,int y);

void bezline(bezier * p,int flatness,bezlinefn fn);

void bezline2(bezier * p,int flatness,bezlinefn fn,int incx,int incy);


*/


typedef struct bezcachestr
{
 bezier  b;
 int   * line;
 int     npoints;



} bezcachestr;



static void * bezhandle;

static int    bezprobe;


// prime numbers

#define BEZCACHEMAX 123
#define BEZPROBEINC 71


#define BEZCHUNK (BEZCACHEMAX*sizeof(bezcachestr))


static int bezcmp(void * xb1,void * xb2)
{
 return(memcmp(xb1,xb2,sizeof(bezier)));
}


static int * bezdatap;
static int   bezdatan;

#define BEZDATACHUNK (sizeof(int)*2*512)

#define BEZFLAT 8


static os_error * bezcurveflat(int x0,int y0,int x1,int y1,void * data)
{
 os_error * err;


 err=schunk((flex_ptr)&bezdatap,(bezdatan+4)*sizeof(int),BEZDATACHUNK);
 if(!err)
 {
  if(bezdatan==0)
  {
   bezdatap[0]=x0;
   bezdatap[1]=y0;
   bezdatan=2;
  }

  bezdatap[bezdatan]=x1;
  bezdatap[bezdatan+1]=y1;
  bezdatan+=2;
 }

 return(err);
}



static os_error * bezcurve2(int ** ptr,int * n,bezier * p)
{
 os_error * err;

 err=NULL;

 bezdatan=0;

 bezflatten(p,BEZFLAT,bezcurveflat,bezdatap);


 *n=bezdatan;
 *ptr=bezdatap;

 return(err);
}



static os_error * bezcurve(int ** ptr,int * n,bezier * p)
{
 os_error    * err;
 bezcachestr * bp;
 int           npoints;
 int           nmax;

 err=NULL;

 bp=(bezcachestr*)xmapsearch(p,bezhandle);
 if(bp)
 {
  *ptr=bp->line;
  *n=bp->npoints;
  cprintf("cache hit");
 }
 else
 {
  nmax=xmapcount(bezhandle);
  if(nmax>BEZCACHEMAX)
  {
   cprintf("cache rem %d",bezprobe);
   bp=(bezcachestr*)xmaprecord(bezhandle,bezprobe);
   bezprobe+=BEZPROBEINC;
   if(bezprobe>=BEZCACHEMAX) bezprobe-=BEZCACHEMAX;
   if(bp)
   {
    sfree((flex_ptr)&bp->line);
    xmaprem(bp,bezhandle);
   }
  }

  err=xmapadd2(p,&bezhandle,(void**)&bp);
  if(!err)
  {
   bezcurve2(ptr,&npoints,p);

   salloc((flex_ptr)&bp->line,sizeof(int)*npoints);
   memcpy(bp->line,*ptr,sizeof(int)*npoints);
   *n=bp->npoints=npoints;
  }
 }

 return(err);
}




os_error * bezinit(void)
{
 os_error * err;

 err=xmapinit(&bezhandle,sizeof(bezcachestr),BEZCHUNK,bezcmp);

 return(err);
}


#ifdef NEVER


/* results of a ray from x,y to left infinity, on a curve cx[0],cx[1]... */
// used to determine if x,y is inside a curve, return number of cuts

int hbintersection(int x,int y,coordstr * cx)
{
 os_error * err;
 int      * ptr;
 coordstr   lx[2];
 int        n;
 int      * lt;
 int        bits;
 int        cuts;

 cuts=0;

 err=bezcurve(&ptr,&n,(bezier*)cx);
 
 lt=ptr+n;

 lx[0].x=*ptr++;
 lx[0].y=*ptr++;

 while(ptr<lt)
 {
  lx[1].x=*ptr++;
  lx[1].y=*ptr++;

  bits=hintersection(x,y,lx);

  if(bits & INTER) cuts++;

  lx[0]=lx[1];
 }


 return(cuts);
}



#endif


/* results of a ray from x,y to left infinity, on a curve cx[0],cx[1]... */
/* in INTERXY X denotes the line, and Y the end of the other line on it  */



#ifdef NEVER

int hbintersection(int x,int y,coordstr * cx)
{
 int         val[2];
 int         uval[2];
 int         i;
 int         n;
 int         roots;
 bezier      p;
 bezsplitstr split[3];



 p.left.y=cx[0].y-y;
 p.leftc.y=cx[1].y-y;
 p.rightc.y=cx[2].y-y;
 p.right.y=cx[3].y-y;

 if((p.left.y^p.leftc.y)<0  || 
    (p.left.y^p.rightc.y)<0 || 
    (p.left.y^p.right.y)<0)
 {
  p.left.x=cx[0].x-x;
  p.leftc.x=cx[1].x-x;
  p.rightc.x=cx[2].x-x;
  p.right.x=cx[3].x-x;

  if((p.left.x^p.leftc.x)<0  || 
     (p.left.x^p.rightc.x)<0 || 
     (p.left.x^p.right.x)<0)
  {
   /* get y roots - calculate x values */

   bezchunks(&p,split,&n,NULL);

 /*  dprintf(10,"get y roots - calculate x values n=%d",n); */

   roots=0;

   for(i=0;i<n;i++)
   {
    roots+=hbsub(&split[i].b);
   }

   return(roots);
  }
  else
  if(p.left.x>0)  /* any y root is an x root */
  {
   roots=0;

   n=bezmaxmin(p.left.y,p.leftc.y,p.rightc.y,p.right.y,val,uval);

  /* dprintf(10,"any y root is an x root n=%d",n); */

   if(n)
   {
    if((p.left.y ^ val[0])<0)      roots++;

    if(n==2)
    {
     if((val[0] ^ val[1])<0)       roots++;
     if((val[1] ^ p.right.y)<0)    roots++;
    }
    else
    {
     if((val[0] ^ p.right.y)<0)    roots++;
    }
   }
   else roots=1;

   return(roots);
  }
 }

 return(0);
}


#endif



#ifdef NEVER

/* intersections of horizontal ray  ray[0]->ray[1] with cx[0],cx[1]... */

int hbintersectionr(coordstr * ray,coordstr * cx,cutstr * result)
{
 int         y;
 int         x;
 int         mx;
 bezier      p;
 int         n;
 int         i;
 int         roots;
 bezsplitstr split[3];

 y=ray[0].y;

 p.left.y  =cx[0].y-y;
 p.leftc.y =cx[1].y-y;
 p.rightc.y=cx[2].y-y;
 p.right.y =cx[3].y-y;

/* if((p.left.y^p.leftc.y)<0  || 
    (p.left.y^p.rightc.y)<0 || 
    (p.left.y^p.right.y)<0)  */

 if(p.left.y>BEZMAXERR   && p.leftc.y>BEZMAXERR && 
    p.rightc.y>BEZMAXERR && p.right.y>BEZMAXERR) return(0);

 if(p.left.y<-BEZMAXERR && p.leftc.y<-BEZMAXERR && 
    p.rightc.y<-BEZMAXERR && p.right.y<-BEZMAXERR) return(0);

 {
  x=ray[0].x;
  mx=ray[1].x-x;

  p.left.x  =cx[0].x-x;
  p.leftc.x =cx[1].x-x;
  p.rightc.x=cx[2].x-x;
  p.right.x =cx[3].x-x;


  if(p.left.x<0  && p.leftc.x<0  && p.rightc.x<0  && p.right.x<0)  return(0);
  if(p.left.x>mx && p.leftc.x>mx && p.rightc.x>mx && p.right.x>mx) return(0);

  roots=0;

  bezchunks((bezier*)cx,split,&n,NULL);

  for(i=0;i<n;i++)
  {

 /*  dprintf(0,"chunk=%d umin=%x umax=%x",i,split[i].umin,split[i].umax);
   dprintf(0,"left x=%d y=%d right x=%d y=%d",
                          split[i].b.left.x,split[i].b.left.y, 
                          split[i].b.right.x,split[i].b.right.y); */


   split[i].b.left.y-=y;
   split[i].b.leftc.y-=y;
   split[i].b.right.y-=y;
   split[i].b.rightc.y-=y;

   split[i].b.left.x-=x;
   split[i].b.leftc.x-=x;
   split[i].b.right.x-=x;
   split[i].b.rightc.x-=x;


   if(hbsubr(&split[i].b,mx,&result[roots],split[i].umin,split[i].umax))
   {
    if(n>1)
    {
     if(i==0)
     {
      if(result[roots].type & INTER01) result[roots].type|=INTERMX2;
      result[roots].type&=~INTER01;
     }
     else
     if(i==(n-1))
     {
      if(result[roots].type & INTER00) result[roots].type|=INTERMX2;
      result[roots].type&=~INTER00;
     }
     else
     {
      if(result[roots].type & INTER00) result[roots].type|=INTERMX2;
      if(result[roots].type & INTER01) result[roots].type|=INTERMX2;
      result[roots].type&=~(INTER00|INTER01);
     }
    }

    result[roots].c.x+=x;
    result[roots].c.y+=y;

   /* dprintf(0,"result at x=%d y=%d",result[roots].c.x,result[roots].c.y); */

    roots++;
   }
  }

  return(roots);
 }

 return(0);
}



int hbintersectionc(coordstr * ray,coordstr * cx,cutstr * result)
{
 int cuts;
 int n;

 cuts=hbintersectionr(ray,cx,result);

 for(n=0;n<cuts;n++)
 {
  if(result[n].type & INTERMX2)
  {
   bezdev((bezier *)cx,result[n].u2,&result[n].dx2,NULL,NULL,&result[n].d2y2);
  }
  else
  {
   bezdev((bezier *)cx,result[n].u2,NULL,&result[n].dy2,NULL,NULL);
  }
 }

 return(cuts);
}


#endif


/****************************************************************************/




#ifdef NEVER

int vbintersectionr(coordstr * ray,coordstr * cx,cutstr * result)
{
 return(bintersection(ray,cx,result));
}



int vbintersectionc(coordstr * ray,coordstr * cx,cutstr * result)
{
 int cuts;
 int n;

 cuts=vbintersectionr(ray,cx,result);

 for(n=0;n<cuts;n++)
 {
  bezdev((bezier *)cx,result[n].u2,&result[n].dx2,&result[n].dy2,NULL,NULL);

  result[n].dx1=ray[1].x-ray[0].x;
  result[n].dy1=ray[1].y-ray[0].y;
 }

 return(cuts);
}


#endif

/****************************************************************************/


#ifdef NEVER


static int      brcount;
static cutstr * bresult;
static int      biumin;
static int      biumax;



#define BEZSPLIT  256
#define LINESPLIT 256


static int bsub(aboxstr * box,bezier * p,int ulo,int uhi)
{
 bezier  a;
 bezier  b;
 aboxstr ba;
 aboxstr bb;
 int     temp;
 int     umid;
 int     splitx;
 int     splity;


 /* box and bezier don't overlap return 0 */

 if(p->left.x<p->right.x)
 {
  ba.x0=p->left.x;
  ba.x1=p->right.x;
 }
 else
 {
  ba.x1=p->left.x;
  ba.x0=p->right.x;
 }


 if(p->leftc.x<ba.x0)  ba.x0=p->leftc.x;
 else
 if(p->leftc.x>ba.x1)  ba.x1=p->leftc.x;

 if(p->rightc.x<ba.x0) ba.x0=p->rightc.x;
 else
 if(p->rightc.x>ba.x1) ba.x1=p->rightc.x;


 if(p->left.y<p->right.y)
 {
  ba.y0=p->left.y;
  ba.y1=p->right.y;
 }
 else
 {
  ba.y1=p->left.y;
  ba.y0=p->right.y;
 }

 if(p->leftc.y<ba.y0)  ba.y0=p->leftc.y;
 else
 if(p->leftc.y>ba.y1)  ba.y1=p->leftc.y;

 if(p->rightc.y<ba.y0) ba.y0=p->rightc.y;
 else
 if(p->rightc.y>ba.y1) ba.y1=p->rightc.y;


 if(box->x0<box->x1)
 {
  if(box->x1<ba.x0) return(0);
  if(box->x0>ba.x1) return(0);
  splitx=(box->x1-box->x0)>LINESPLIT;
 }
 else
 {
  if(box->x0<ba.x0) return(0);
  if(box->x1>ba.x1) return(0);
  splitx=(box->x0-box->x1)>LINESPLIT;
 }

 if(box->y0<box->y1)
 {
  if(box->y1<ba.y0) return(0);
  if(box->y0>ba.y1) return(0);
  splity=(box->y1-box->y0)>LINESPLIT;
 }
 else
 {
  if(box->y0<ba.y0) return(0);
  if(box->y1>ba.y1) return(0);
  splity=(box->y0-box->y1)>LINESPLIT;
 }


 if((ba.x1-ba.x0)<=BEZSPLIT && (ba.y1-ba.y0)<=BEZSPLIT)
 {
  coordstr cx1[2];
  coordstr cx2[2];
  int      code;

  cx1[0].x=box->x0;
  cx1[0].y=box->y0;
  cx1[1].x=box->x1;
  cx1[1].y=box->y1;

  cx2[0].x=p->left.x;
  cx2[0].y=p->left.y;
  cx2[1].x=p->right.x;
  cx2[1].y=p->right.y;

  code=intersection(cx1,cx2,&bresult[brcount]);
  if(code)
  {
   if(!(code & (INTER00|INTER10|INTERCO)))
   {
    if((ba.x1-ba.x0)>(ba.y1-ba.y0))
    {
     if(cx2[0].x==ba.x0)
     {
      bresult[brcount].u2=ulo+scale(bresult[brcount].c.x-ba.x0,uhi-ulo,
                                                   ba.x1-ba.x0);
     }
     else
     {
      bresult[brcount].u2=ulo+scale(ba.x1-bresult[brcount].c.x,uhi-ulo,
                                    ba.x1-ba.x0);
     }
    }
    else
    {
     if(cx2[0].y==ba.y0)
     {
      bresult[brcount].u2=ulo+scale(bresult[brcount].c.y-ba.y0,uhi-ulo,
                                                   ba.y1-ba.y0);
     }
     else
     {
      bresult[brcount].u2=ulo+scale(ba.y1-bresult[brcount].c.y,uhi-ulo,
                                    ba.y1-ba.y0);
     }
    }

    bresult[brcount].umin2=biumin;
    bresult[brcount].umax2=biumax;
    brcount++;
   }
  }

  return(1);
 }


 a.left=p->left;
 b.right=p->right;

  /* x coords */

 temp=(p->leftc.x+p->rightc.x)/2;

 a.leftc.x =(p->left.x+p->leftc.x)/2;
 a.rightc.x=(temp+a.leftc.x)/2;
 b.rightc.x=(p->rightc.x+p->right.x)/2;
 b.leftc.x =(temp+b.rightc.x)/2;

 a.right.x=(a.rightc.x+b.leftc.x)/2;

  /* y coords */

 temp=(p->leftc.y+p->rightc.y)/2;

 a.leftc.y =(p->left.y+p->leftc.y)/2;
 a.rightc.y=(temp+a.leftc.y)/2;
 b.rightc.y=(p->rightc.y+p->right.y)/2;
 b.leftc.y =(temp+b.rightc.y)/2;

 a.right.y=(a.rightc.y+b.leftc.y)/2;

 b.left=a.right;

 umid=(uhi+ulo)/2;

 if(splitx || splity)
 {
  ba.x0=box->x0;
  ba.x1=(box->x1+box->x0)/2;
  bb.x0=ba.x1; 
  bb.x1=box->x1;

  ba.y0=box->y0;
  ba.y1=(box->y0+box->y1)/2;
  bb.y0=ba.y1;
  bb.y1=box->y1;

  bsub(&ba,&a,ulo,umid);
  bsub(&ba,&b,umid,uhi);

  bsub(&bb,&a,ulo,umid);
  bsub(&bb,&b,umid,uhi);
 }
 else
 {
  bsub(box,&a,ulo,umid);
  bsub(box,&b,umid,uhi);
 }

 return(0);
}





/*

 get curvature


 int     x1;
 int     x2;
 int     y1;
 int     y2;
 int     x;
 int     y;
 int     l;


 x1=(coord1[2].x-coord1[1].x)-(coord1[1].x-coord1[0].x);
 if(x1<0) x1=-x1;
 x2=(coord1[3].x-coord1[2].x)-(coord1[2].x-coord1[1].x);
 if(x2<0) x2=-x2;

 y1=(coord1[2].y-coord1[1].y)-(coord1[1].y-coord1[0].y);
 if(y1<0) y1=-y1;
 y2=(coord1[3].y-coord1[2].y)-(coord1[2].y-coord1[1].y);
 if(y2<0) y2=-y2;

 x=MAX(x1,x2);
 y=MAX(y1,y2);

 l=MAX(x,y);

*/



/* coord0 2 points of line  */
/* coord1 4 points of curve */


int bintersection(coordstr * coord0,coordstr * coord1,cutstr * result)
{
 aboxstr     box;
 int         i;
 int         n;
 bezsplitstr split[3];
 int         dx1;
 int         dy1;

 box.x0=coord0[0].x;
 box.y0=coord0[0].y;
 box.x1=coord0[1].x;
 box.y1=coord0[1].y;

 brcount=0;       /* this is a global static variable */
 bresult=result;  /* this is a global static variable */

 bezchunks((bezier*)coord1,split,&n,NULL);
 for(i=0;i<n;i++)
 {
  biumin=split[i].umin;
  biumax=split[i].umax;
  bsub(&box,&split[i].b,biumin,biumax);
 }

 if(brcount)
 {
  dx1=box.x1-box.x0;
  dy1=box.y1-box.y0;

  for(i=0;i<brcount;i++)
  {
   result[i].dx1=dx1;
   result[i].dy1=dy1;
  }
 }

 return(brcount);
}







static void bqbound(bezier * p,aboxstr * ba)
{
 register int lo;
 register int hi;

 if(p->left.x<p->right.x)
 {
  lo=p->left.x;
  hi=p->right.x;
 }
 else
 {
  lo=p->right.x;
  hi=p->left.x;
 }

 if(p->leftc.x<lo)  lo=p->leftc.x;
 else
 if(p->leftc.x>hi)  hi=p->leftc.x;

 if(p->rightc.x<lo) lo=p->rightc.x;
 else
 if(p->rightc.x>hi) hi=p->rightc.x;

 ba->x0=lo;
 ba->x1=hi;

 if(p->left.y<p->right.y)
 {
  lo=p->left.y;
  hi=p->right.y;
 }
 else
 {
  lo=p->right.y;
  hi=p->left.y;
 }

 if(p->leftc.y<lo)  lo=p->leftc.y;
 else
 if(p->leftc.y>hi)  hi=p->leftc.y;

 if(p->rightc.y<lo) lo=p->rightc.y;
 else
 if(p->rightc.y>hi) hi=p->rightc.y;

 ba->y0=lo;
 ba->y1=hi;
}


static void bdivide(bezier * p,bezier * a,bezier * b)
{
 int temp;

 a->left=p->left;
 b->right=p->right;

  /* x coords */

 temp=(p->leftc.x+p->rightc.x)/2;

 a->leftc.x =(p->left.x+p->leftc.x)/2;
 a->rightc.x=(temp+a->leftc.x)/2;
 b->rightc.x=(p->rightc.x+p->right.x)/2;
 b->leftc.x =(temp+b->rightc.x)/2;

 a->right.x=(a->rightc.x+b->leftc.x)/2;

  /* y coords */

 temp=(p->leftc.y+p->rightc.y)/2;

 a->leftc.y =(p->left.y+p->leftc.y)/2;
 a->rightc.y=(temp+a->leftc.y)/2;
 b->rightc.y=(p->rightc.y+p->right.y)/2;
 b->leftc.y =(temp+b->rightc.y)/2;

 a->right.y=(a->rightc.y+b->leftc.y)/2;

 b->left=a->right;
}




int bezvalx(int p0,int p1,int p2,int p3,int u)
{
 int temp;
 int up;
 int aleftc;
 int arightc;
 int bleftc;
 int brightc;


 u=CUT(u,0,0x10000);

 up=0x10000-u;

 temp=ttmul(up,p1)+ttmul(u,p2);

 aleftc =ttmul(up,p0)+ttmul(u,p1);
 arightc=ttmul(up,aleftc)+ttmul(u,temp);

 brightc=ttmul(up,p2)+ttmul(u,p3);
 bleftc =ttmul(up,temp)+ttmul(u,brightc);

 return(ttmul(up,arightc)+ttmul(u,bleftc));
}





/* 0<=u<=0x10000 */

void bezsplit2(bezier * p,bezier * a,bezier * b,int u)
{
 int temp;
 int up;

 u=CUT(u,0,0x10000);

 if(u==0)
 {
  *b=*p;
  a->left=p->left;
  a->leftc=p->left;
  a->rightc=p->left;
  a->right=p->left;
 }
 else
 {
  up=0x10000-u;

  a->left=p->left;
  b->right=p->right;

  /* x coords */

  temp=ttmul(up,p->leftc.x)+ttmul(u,p->rightc.x);

  a->leftc.x =ttmul(up,p->left.x)+ttmul(u,p->leftc.x);
  a->rightc.x=ttmul(up,a->leftc.x)+ttmul(u,temp);
  b->rightc.x=ttmul(up,p->rightc.x)+ttmul(u,p->right.x);
  b->leftc.x =ttmul(up,temp)+ttmul(u,b->rightc.x);

  a->right.x=ttmul(up,a->rightc.x)+ttmul(u,b->leftc.x);

  /* y coords */

  temp=ttmul(up,p->leftc.y)+ttmul(u,p->rightc.y);

  a->leftc.y =ttmul(up,p->left.y)+ttmul(u,p->leftc.y);
  a->rightc.y=ttmul(up,a->leftc.y)+ttmul(u,temp);
  b->rightc.y=ttmul(up,p->rightc.y)+ttmul(u,p->right.y);
  b->leftc.y =ttmul(up,temp)+ttmul(u,b->rightc.y);

  a->right.y=ttmul(up,a->rightc.y)+ttmul(u,b->leftc.y);

  b->left=a->right;
 }
}






void bezsplit3(bezier * p,bezier * a,bezier * b,bezier * c,int u1,int u2)
{
 bezier t;

 u1=CUT(u1,0,0x10000);
 u2=CUT(u2,0,0x10000);

 if(u1==0)
 {
  a->left=p->left;
  a->leftc=p->left;
  a->rightc=p->left;
  a->right=p->left;

  bezsplit2(p,b,c,u2);
 }
 else
 if(u2==0x10000)
 {
  c->left=p->right;
  c->leftc=p->right;
  c->rightc=p->right;
  c->right=p->right;

  bezsplit2(p,a,b,u1);
 }
 else
 {
  if(u1<0x8000)
  {
   bezsplit2(p,a,&t,u1);
   u2=scale(u2-u1,0x10000,0x10000-u1);
   bezsplit2(&t,b,c,u2);
  }
  else
  {
   bezsplit2(p,&t,c,u2);
   u1=scale(u1,0x10000,u2);
   bezsplit2(&t,a,b,u1);
  }
 }
}
















static int biumin1;
static int biumin2;
static int biumax1;
static int biumax2;



static int bbsub(bezier * p1,int ulo1,int uhi1,
                 bezier * p2,int ulo2,int uhi2)
{
 bezier  a1;
 bezier  b1;
 aboxstr ba1;
 bezier  a2;
 bezier  b2;
 aboxstr ba2;
 int     dx1;
 int     dy1;
 int     dx2;
 int     dy2;
 int     umid1;
 int     umid2;
 int     split1;
 int     split2;


 /* don't overlap return 0 */

 bqbound(p1,&ba1);
 bqbound(p2,&ba2);

 if(ba1.x0>ba2.x1 || ba1.x1<ba2.x0) return(0);
 if(ba1.y0>ba2.y1 || ba1.y1<ba2.y0) return(0);

 umid1=(ulo1+uhi1)/2;
 umid2=(ulo2+uhi2)/2;

 dx1=ba1.x1-ba1.x0;
 dx2=ba2.x1-ba2.x0;
 dy1=ba1.y1-ba1.y0;
 dy2=ba2.y1-ba2.y0;

 split1=(dx1>BEZSPLIT) || (dy1>BEZSPLIT);
 split2=(dx2>BEZSPLIT) || (dy2>BEZSPLIT);

 if(!split1 && !split2)
 {
  coordstr cx1[2];
  coordstr cx2[2];
  int      code;

  cx1[0].x=p1->left.x;
  cx1[0].y=p1->left.y;
  cx1[1].x=p1->right.x;
  cx1[1].y=p1->right.y;

  cx2[0].x=p2->left.x;
  cx2[0].y=p2->left.y;
  cx2[1].x=p2->right.x;
  cx2[1].y=p2->right.y;

  code=intersection(cx1,cx2,&bresult[brcount]);
  if(code)
  {
   if(!(code & (INTER00|INTER10|INTERCO)))
   {
    if((ba1.x1-ba1.x0)>(ba1.y1-ba1.y0))
    {
     if(cx1[0].x==ba1.x0)
     {
      bresult[brcount].u1=ulo1+scale(bresult[brcount].c.x-ba1.x0,uhi1-ulo1,
                                                   ba1.x1-ba1.x0);
     }
     else
     {
      bresult[brcount].u1=ulo1+scale(ba1.x1-bresult[brcount].c.x,uhi1-ulo1,
                                     ba1.x1-ba1.x0);
     }
    }
    else
    {
     if(cx1[0].y==ba1.y0)
     {
      bresult[brcount].u1=ulo1+scale(bresult[brcount].c.y-ba1.y0,uhi1-ulo1,
                                                   ba1.y1-ba1.y0);
     }
     else
     {
      bresult[brcount].u1=ulo1+scale(ba1.y1-bresult[brcount].c.y,uhi1-ulo1,
                                     ba1.y1-ba1.y0);
     }
    }

    if((ba2.x1-ba2.x0)>(ba2.y1-ba2.y0))
    {
     if(cx2[0].x==ba2.x0)
     {
      bresult[brcount].u2=ulo2+scale(bresult[brcount].c.x-ba2.x0,uhi2-ulo2,
                                                   ba2.x1-ba2.x0);
     }
     else
     {
      bresult[brcount].u2=ulo2+scale(ba2.x1-bresult[brcount].c.x,uhi2-ulo2,
                                     ba2.x1-ba2.x0);
     }
    }
    else
    {
     if(cx2[0].y==ba2.y0)
     {
      bresult[brcount].u2=ulo2+scale(bresult[brcount].c.y-ba2.y0,uhi2-ulo2,
                                                   ba2.y1-ba2.y0);
     }
     else
     {
      bresult[brcount].u2=ulo2+scale(ba2.y1-bresult[brcount].c.y,uhi2-ulo2,
                                     ba2.y1-ba2.y0);
     }
    }


    bresult[brcount].umin1=biumin1;
    bresult[brcount].umax1=biumax1;
    bresult[brcount].umin2=biumin2;
    bresult[brcount].umax2=biumax2;

    brcount++;
   }
  }

  return(1);
 }

 if(split1 && split2)
 {
  bdivide(p1,&a1,&b1);
  bdivide(p2,&a2,&b2);

  bbsub(&a1,ulo1,umid1,&a2,ulo2,umid2);
  bbsub(&a1,ulo1,umid1,&b2,umid2,uhi2);

  bbsub(&b1,umid1,uhi1,&a2,ulo2,umid2);
  bbsub(&b1,umid1,uhi1,&b2,umid2,uhi2);
 }
 else
 if(split1)
 {
  bdivide(p1,&a1,&b1);
  bbsub(&a1,ulo1,umid1,p2,ulo2,uhi2);
  bbsub(&b1,umid1,uhi1,p2,ulo2,uhi2);
 }
 else
 /*if(split2) */
 {
  bdivide(p2,&a2,&b2);
  bbsub(p1,ulo1,uhi1,&a2,ulo2,umid2);
  bbsub(p1,ulo1,uhi1,&b2,umid2,uhi2);
 }

 return(0);
}



/* coord0 4 points of curve */
/* coord1 4 points of curve */


int bbintersection(coordstr * coord0,coordstr * coord1,cutstr * result)
{
 int         i;
 int         j;
 int         n1;
 int         n2;
 bezsplitstr split1[3];
 bezsplitstr split2[3];

 brcount=0;
 bresult=result;

 bezchunks((bezier*)coord0,split1,&n1,NULL);
 bezchunks((bezier*)coord1,split2,&n2,NULL);

 for(i=0;i<n1;i++)
 {
  for(j=0;j<n2;j++)
  {
   biumin1=split1[i].umin;
   biumax1=split1[i].umax;
   biumin2=split2[j].umin;
   biumax2=split2[j].umax;

   bbsub(&split1[i].b,split1[i].umin,split1[i].umax,
         &split2[j].b,split2[j].umin,split2[j].umax);
  }
 }

 return(brcount);
}





void bezval(bezier * b,coordstr * c,int u)
{
 c->x=bezvalx(b->left.x,b->leftc.x,b->rightc.x,b->right.x,u);
 c->y=bezvalx(b->left.y,b->leftc.y,b->rightc.y,b->right.y,u);
}


#endif




void bezgetext(bezier * b,aboxstr * abox)
{
 os_error * err;
 int      * ptr;
 int        n;
 int      * lt;
 int        x;
 int        y;




 err=bezcurve(&ptr,&n,b);
 
 lt=ptr+n;
 abox->x0=abox->x1=*ptr++;
 abox->y0=abox->y1=*ptr++;

 while(ptr<lt)
 {
  x=*ptr++;
  y=*ptr++;

  if(abox->x0>x) abox->x0=x;
  if(abox->x1<x) abox->x1=x;
  if(abox->y0>y) abox->y0=y;
  if(abox->y1<y) abox->y1=y;
 }

 cprintf("bezgetext x0 %d x1 %d y0 %d y1%d n=%d",abox->x0,abox->x1,abox->y0,abox->y1,n);
}





#ifdef NEVER



void bezdev(bezier * b,int u,int * dx,int * dy,int * d2x,int * d2y)
{
 int up;
 int u2;
 int up2;
 int uup;

 up=0x10000-u;
 u2=ttmul(u,u);
 up2=ttmul(up,up);
 uup=ttmul(up,u);

 if(dx)   *dx=ttmul(b->left.x,-up2)+
              ttmul(b->leftc.x,-2*uup+up2)+
              ttmul(b->rightc.x,-u2+2*uup)+
              ttmul(b->right.x,u2);
 
 if(d2x) *d2x=ttmul(b->left.x,2*up)+
              ttmul(b->leftc.x,2*u-4*up)+
              ttmul(b->rightc.x,-4*u+2*up)+
              ttmul(b->right.x,2*u);

 if(dy)   *dy=ttmul(b->left.y,-up2)+
              ttmul(b->leftc.y,-2*uup+up2)+
              ttmul(b->rightc.y,-u2+2*uup)+
              ttmul(b->right.y,u2);
 
 if(d2y) *d2y=ttmul(b->left.y,2*up)+
              ttmul(b->leftc.y,2*u-4*up)+
              ttmul(b->rightc.y,-4*u+2*up)+
              ttmul(b->right.y,2*u);
}


#endif

#endif