/*->c.frameflow */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "font.h"
#include "transform.h"
#include "xprocess.h"


#include "err.h"
#include "xmath.h"
#include "fsx.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "colour.h"
#include "etc.h"
#include "deb.h"


#include "constants.h"
#include "str.h"
#include "object.h"

#include "view.h"
#include "viewr.h"
#include "file.h"

#include "xredraw.h"
#include "curve.h"
#include "posn.h"

#include "chain.h"


#include "txstr.h"
#include "fl.h"


#include "frameflow.h"


/* 

   a short module, concerned with generating the flows needed when
   a frame is modified 

*/




typedef struct
{
 int object;
 int blanked;

} frliststr;


#define FRCHUNK 32*sizeof(frliststr)


static frliststr * frlist;
static int         frentries;



os_error * frflstart(filestr * file)
{
 os_error * err;

 if(!frlist) err=flex_alloc((flex_ptr)&frlist,FRCHUNK);
 else        err=flex_chunk((flex_ptr)&frlist,FRCHUNK,FRCHUNK);

 frentries=0;

 return(err);

 USE(file);
}



static int frcmp(const void * p1,const void * p2)
{
 return(((frliststr*)p1)->object-((frliststr*)p2)->object);
}


os_error * frflblanked(filestr * file,int object,int * blanked)
{
 frliststr   key;
 frliststr * hit;

 *blanked=0;

 if(frlist && frentries)
 {
  key.object=object;
  hit=(frliststr*)bsearch(&key,frlist,frentries,sizeof(frliststr),frcmp);
  if(hit) *blanked=hit->blanked;
 }

 return(NULL);
 USE(file);
}



os_error * frflflush(filestr * file)
{
 os_error  * err;
 int         r;
 int         w;
 objhdrstr * obj;

 err=NULL;


// cprintf("frflflush in ");

 if(frlist)
 {
  if(frentries)
  {
   qsort(frlist,frentries,sizeof(frliststr),frcmp);
   r=0;
   w=0;
   for(r=0;r<frentries;r++)
   {
    if(w==0 || frlist[r].object!=frlist[w-1].object)
    {
     frlist[w]=frlist[r];
     w++;
    }
   }

   frentries=w;

   for(r=0;r<w;r++)
   {
    obj=(objhdrstr*)fob_find(frlist[r].object,file);
    objectreflow(obj,file);
   }
  }
 }


// cprintf("frflflush exit");

 return(err);
}



os_error * frflscrap(filestr * file)
{
// if(frentries) cprintf("scrap fl entries");
 frentries=0;
 return(flex_free((flex_ptr)&frlist));
 USE(file);
}



os_error * frflend(filestr * file)
{
 os_error  * err;

 err=frflflush(file);
 frflscrap(file);

 return(err);
}




os_error * frflvalid(filestr * file)
{
 os_error * err;
 int        i;

 err=NULL;

 for(i=0;i<frentries;i++)
 {
  err=flowvalidbox(frlist[i].object,file);
 }

 return(err);
}



/* called by objects that intersect with the object that has changed */

os_error * frfladdb(sobjectstr ** sob,filestr * file,int blanked)
{
 os_error * err;
 int        key;

 err=NULL;

 if(frlist)
 {
  key=(*sob)->key;

  err=flex_chunk((flex_ptr)&frlist,(frentries+1)*sizeof(frliststr),FRCHUNK);
  if(!err)
  {
   frlist[frentries].blanked=blanked;
   frlist[frentries++].object=key;
  }

  *sob=(sobjectstr*)fob_find(key,file);
 }

 return(err);

 USE(file);
}


os_error * frfladd(sobjectstr ** sob,filestr * file)
{
 return(frfladdb(sob,file,0));
}



static sobjectstr * firstopaque(sobjectstr * start,filestr * file)
{
 objectflowstr tf;
 int           parent;

 while(1)
 {
  parent=parentof((chainobject*)start,file);
  if(!parent) break;

  start=(sobjectstr*)fob_find(parent,file);
  objectflow((objhdrstr*)start,file,&tf,1);

  if(tf.flow.omode==FROFRAME)
  {
   return(start);
  }
 }

 return(NULL);
}



/* means reflow all objects which this object intersects with */

os_error * frameflow(sobjectstr ** sob,filestr * file)
{
 os_error   * err;
 sobjectstr * start;
 sobjectstr   sobfix;
 int          end;

 err=NULL;

/* dprintf(0,"frameflow"); */


 if(frlist)
 {
  sobfix=**sob;

  if(embededobject(sobfix.key,file))
  {
			if(sobfix.chain.parent)				// if this function is applied to a text frame that has been deleted it is unlinked, no parent, no next or prev
			{
    start=(sobjectstr*)fob_find(sobfix.chain.parent,file);
    end=2;
			}
			else
			{
				start=NULL;
				end=0;
			}
  }
  else
  {
   start=firstopaque(&sobfix,file);
   end=0;
  }

  if(start)
  {
   err=objectsetflow((objhdrstr*)start,(objhdrstr*)&sobfix,file,&end);
  }

  *sob=(sobjectstr*)fob_find(sobfix.key,file);
 }

 return(err);
}


/* if we are deleting the object, remove any refs from flow list */

os_error * frameflowdelete(sobjectstr ** sob,filestr * file)
{
 os_error * err;
 int        i;
 int        key;

 err=frameflow(sob,file);
 if(!err)
 {
  key=(*sob)->key;

  if(frlist && frentries)
  {
   for(i=0;i<frentries;i++)
   {
    if(frlist[i].object==key)
    {
     frentries--;
     frlist[i]=frlist[frentries];
     i--;
    }
   }
  }
 }

 return(err);
}


