/*->c.frametool */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "akbd.h"
#include "transform.h"


#include "err.h"
#include "fsx.h"
#include "bf.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "etc.h"
#include "colour.h"
#include "key.h"
#include "xmath.h"
#include "deb.h"
#include "dbhi.h"

#include "str.h"
#include "constants.h"
#include "mhi.h"

#include "object.h"
#include "chain.h"

#include "reslink.h"

#include "xredraw.h"
#include "curvelo.h"
#include "curve.h"
#include "geometry.h"

#include "viewo.h"
#include "viewr.h"
#include "view.h"
#include "file.h"
#include "caret.h"
#include "select.h"
#include "posn.h"
#include "pref.h"
#include "rulers.h"

#include "undo.h"
#include "frameflow.h"

#include "units.h"
#include "xcolour.h"
#include "objectdb.h"
#include "grid.h"
#include "gline.h"

#include "infopal.h"
#include "editpath.h"
#include "toolstr.h"
#include "tools.h"

#include "frametool.h"



#define UCREATES 0x9000
#define UCREATEF 0x9001






static pointstr drawptr=
{
 "p_draw",6,6
};




/*****************************************************************************/
/* Frame tool */


typedef struct frametoolstr
{
 posnstr  framestart;
 posnstr  framelast;
 coordstr seg;
 int      framestate;
 int      framesquare;
 int      frametool;
 int      framecurtool;     /* what tool was for redraws etc. */
 int      framemode;
 int      framediagramtool;

 int  *   framepath;
 int      framepathptr;     /* pointer to last element */
 int      framepathptr1;    /* pointer to next to last element */
 int      framepathptr2;    /* pointer to next to next to last element */
 int      framepathptr3;    /* pointer to next to next ...element */
 int      framepathptr4;    /* pointer to next to next ...element */
 int      framepathcatch;   /* how many points to go back              */

} frametoolstr;


static int framedrag;


#define FRAMECHUNK    48*sizeof(int)

#define FRAME_NONE   -1

#define FRAME_RECT    TL_RECTANGLE      /* types of frame tool */
#define FRAME_ELLIPSE TL_ELLIPSE
#define FRAME_STRSEG  TL_STRSEG
#define FRAME_CURSEG  TL_CURSEG


#define FRAME_FRAME   1      /* mode for frame code */
#define FRAME_DIAGRAM 2



static void calctansub(int * result,int * prev,int * curr,
                                                 int * next,int forward)
{
 double prevangle;
 double nextangle;
 double diff;
 double normal;
 double tangent;
 int    prevdist;
 int    nextdist;


 if(prev) 
 {
  prevdist=pythag(curr[0]-prev[0],curr[1]-prev[1])/3;
  if(prevdist)
         prevangle=atan2((double)prev[1]-(double)curr[1],
                         (double)prev[0]-(double)curr[0]);
  else   prevangle=0.0;
 }
 else
 {
  prevangle=0.0;
  prevdist=0;
 }

 if(next)
 {
  nextdist=pythag(curr[0]-next[0],curr[1]-next[1])/3;
  if(nextdist)
         nextangle=atan2((double)next[1]-(double)curr[1],
                         (double)next[0]-(double)curr[0]);
  else   nextangle=0.0;
 }
 else
 {
  nextangle=0.0;
  nextdist=0;
 }


 if(!prev) prevangle=nextangle+PI;
 if(!next) nextangle=prevangle-PI;


 diff=prevangle-nextangle;
 normal=prevangle-diff/2;

 if(diff<0) normal=normal+PI;


 if(forward)
 {
  tangent=normal+PI/2;
  result[0]=curr[0]+(int)(prevdist*cos(tangent));
  result[1]=curr[1]+(int)(prevdist*sin(tangent));
 }
 else
 {
  tangent=normal-PI/2;
  result[0]=curr[0]+(int)(nextdist*cos(tangent));
  result[1]=curr[1]+(int)(nextdist*sin(tangent));
 }
}


static void calctangents(int * path,int skip,int doloop)
{
 int * curr;
 int * prev;
 int * prevp;
 int * next;
 int * first;
 int   looped;
 int   loop;

 prevp=NULL;
 prev=NULL;
 curr=NULL;

 first=path;
 looped=0;
 loop=doloop;



 while((*path && !looped) || ((looped || !*path) && doloop && loop--))
 {
  if(*path==DRAWEND || *path==DRAWCLOSE)
  {
   looped=1;
   path=first+3;
  }

  if(*path==DRAWMOVE || *path==DRAWDRAW)
  {
   curr=path+1;
   path+=3;
   skip--;
  }
  else
  if(*path==DRAWCURVE)   /* need last point and next */
  {
   if(path[7]==DRAWEND || path[7]==DRAWCLOSE)
   {
    if(doloop)
    {
          /* assume it starts with a move */

     if(first[3]==DRAWMOVE || first[3]==DRAWDRAW)
     {
      next=first+(3+1);
     }
     else
     if(first[3]==DRAWCURVE) next=first+(3+5);
     else                    next=NULL;
    }
    else next=NULL;
   }
   else
   if(path[7]==DRAWMOVE || path[7]==DRAWDRAW)
   {
    next=path+(7+1);
   }
   else
   if(path[7]==DRAWCURVE)
   {
    next=path+(7+5);
   }
   else
   {
    next=NULL;
   }

   curr=path+5;

   if(skip--<=0)
   {
    calctansub(path+1,prevp,prev,curr,0);
    calctansub(path+3,prev,curr,next,1);
   }

   path+=7;
  }


  prevp=prev;
  prev=curr;
 }
}




static os_error * blob(int x,int y,viewredrawstr * v,int colour)
{
 os_error * err;

 err=rseteorcol(WIMPWHITE,colour);

 x=v->ox+osunits(v->xpo+x,v->view);  
 y=v->oy-osunits(v->ypo+y,v->view);  

 err=bbc_move(x+DRSUB/2,y+DRSUB/2);
 err=bbc_plot(96+5,x-DRSUB/2,y-DRSUB/2);

 return(err);
}






#define NOTLASTBLOB    0x1
#define NOTFIRSTBLOB   0x2
#define LASTTANGENT    0x4
#define BLOBSELECTION  0x8




static os_error * rpatheditline(int * path,viewredrawstr * v,int flags)
{
 os_error * err;
 int      * p;
 int        first;
 int        x;
 int        y;
 int        x1;
 int        x2;
 int        x3;
 int        y1;
 int        y2;
 int        y3;


 err=NULL;

 p=path;
 first=1;
 x=y=0;           /* compiler */


 while(*p)
 {
  if(*p==DRAWCLOSE)
  {
   p++;
  }
  else
  if(*p==DRAWMOVE || *p==DRAWDRAW)
  {
   x=p[1];
   y=p[2];

   if((flags & NOTLASTBLOB) && !p[3]) break;

   if(!((flags & NOTFIRSTBLOB) && first))
   {
    err=blob(x,y,v,WIMPLBLUE);
   }

   p+=3;
  }
  else
  if(*p==DRAWCURVE)
  {

   /* if point after this point is the last, blob the second point,
      else
      if the this point is the last blob the first                */

   x1=*(p+1);
   y1=*(p+2);
   x2=*(p+3);
   y2=*(p+4);
   x3=*(p+5);
   y3=*(p+6);


   if(!(flags & NOTLASTBLOB) || p[7])
   {
    if(!((flags & NOTFIRSTBLOB) && first))
    {
     err=blob(x3,y3,v,WIMPLBLUE);
    }
   }


   if(flags & LASTTANGENT)
   {
    if(!p[7])
    {
     rseteorcol(WIMPWHITE,WIMPGREY2);
     rbbc_move(x,y,v);
     rbbc_draw(x1,y1,v);
     err=blob(x1,y1,v,WIMPORANGE);
    }
    else
    if(((p[7]==DRAWMOVE || p[7]==DRAWDRAW) && !p[7+3]) ||
       (p[7]==DRAWCURVE && !p[7+7]))
    {
     rseteorcol(WIMPWHITE,WIMPGREY2);
     rbbc_move(p[3],p[4],v);
     rbbc_draw(p[5],p[6],v);
     err=blob(p[3],p[4],v,WIMPORANGE);
    }
   }

   x=x3;
   y=y3;

   p+=7;
  }

  first=0;
 }

 return(err);
}




static os_error * framepathline(int * path,viewredrawstr * v,int flags)
{
 os_error * err;
 rsimpleline(path,v,NULL);
 err=rpatheditline(path,v,flags);
 return(err);
}



static os_error * framepathlinepts(frametoolstr * ftool,
                                int pathptr,viewredrawstr * v,int flags)
{
 os_error * err;
 int        temp;

 if(ftool->framepath[pathptr]==DRAWDRAW)
 {
  ftool->framepath[pathptr]=DRAWMOVE;
  err=framepathline(ftool->framepath+pathptr,v,flags);
  ftool->framepath[pathptr]=DRAWDRAW;
 }
 else
 if(ftool->framepath[pathptr]==DRAWMOVE)
 {
  err=framepathline(ftool->framepath+pathptr,v,flags);
 }
 else
 if(ftool->framepath[pathptr]==DRAWCURVE)
 {
  temp=ftool->framepath[pathptr+4];
  ftool->framepath[pathptr+4]=DRAWMOVE;
  err=framepathline(ftool->framepath+pathptr+4,v,flags);
  ftool->framepath[pathptr+4]=temp;
 }
 else err=NULL;

 return(err);
}



static os_error * frameredraw(toolboxstr * toolbox,viewredrawstr * v)
{
 os_error     * err;
 frametoolstr * ftool;
 int            lastblob;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 err=NULL;

 if(ftool->framestate && v->page==ftool->framestart.page)
 {
  if(ftool->framecurtool==FRAME_RECT)
  {
   rseteorcol(WIMPWHITE,WIMPGREY2);
   bbc_rectangle(
           v->ox+osunits(ftool->framestart.xpo+ftool->framestart.x,v->view),
           v->oy-osunits(ftool->framestart.ypo+ftool->framestart.y,v->view),
               osunits(ftool->framelast.x-ftool->framestart.x,v->view),
               osunits(ftool->framestart.y-ftool->framelast.y,v->view));
  }
  else
  if(ftool->framecurtool==FRAME_ELLIPSE)
  {
   rseteorcol(WIMPWHITE,WIMPGREY2);
   bbc_move(v->ox+osunits(ftool->framestart.xpo+ftool->framestart.x,v->view),
            v->oy-osunits(ftool->framestart.ypo+ftool->framestart.y,v->view));

   bbc_move(v->ox+osunits(ftool->framelast.xpo+ftool->framelast.x,v->view),
            v->oy-osunits(ftool->framestart.ypo+ftool->framestart.y,v->view));

   bbc_plot(bbc_Ellipse+5,
           v->ox+osunits(ftool->framestart.xpo+ftool->framestart.x,v->view),
           v->oy-osunits(ftool->framelast.ypo+ftool->framelast.y,v->view));
  }
  else
  if(ftool->framecurtool==FRAME_STRSEG || ftool->framecurtool==FRAME_CURSEG)
  {
   rseteorcol(WIMPWHITE,WIMPGREY2);

   if(framedrag==1) lastblob=0;
   else             lastblob=NOTLASTBLOB;

   if(ftool->framepathcatch==0)
   {
    err=framepathline(ftool->framepath,v,lastblob|LASTTANGENT);
   }
   else
   if(ftool->framepathcatch==1)
   {
    framepathlinepts(ftool,ftool->framepathptr1,v,
                                       NOTFIRSTBLOB|lastblob|LASTTANGENT);
   }
   else
   if(ftool->framepathcatch==2)
   {
    framepathlinepts(ftool,ftool->framepathptr2,v,
                                       NOTFIRSTBLOB|lastblob|LASTTANGENT);
   }
   else
   if(ftool->framepathcatch==3)
   {
    framepathlinepts(ftool,ftool->framepathptr3,v,
                                       NOTFIRSTBLOB|lastblob|LASTTANGENT);
   }
   else
   if(ftool->framepathcatch==4)
   {
    framepathlinepts(ftool,ftool->framepathptr1,v,lastblob|LASTTANGENT);
   }
  }
 }
 return(NULL);
}


/* logic is that first click of double click created a spare point */

static os_error * frameclose(frametoolstr * ftool,int closed)
{
 os_error * err;

 err=flex_chunk((flex_ptr)&ftool->framepath,
                (ftool->framepathptr+12)*sizeof(int),FRAMECHUNK);

 if(closed)
 {
  if(ftool->frametool==FRAME_STRSEG)
  {
   ftool->framepath[ftool->framepathptr+1]=ftool->framestart.x;
   ftool->framepath[ftool->framepathptr+2]=ftool->framestart.y;
   ftool->framepathptr+=3;
  }
  else
  if(ftool->frametool==FRAME_CURSEG)
  {
   ftool->framepath[ftool->framepathptr+5]=ftool->framestart.x;
   ftool->framepath[ftool->framepathptr+6]=ftool->framestart.y;
   ftool->framepathptr+=7;
  }

  ftool->framepath[ftool->framepathptr++]=DRAWCLOSE;
  ftool->framepath[ftool->framepathptr++]=DRAWEND;

  calctangents(ftool->framepath,0,1);
 }
 else
 {
  ftool->framepath[ftool->framepathptr++]=DRAWEND;
  calctangents(ftool->framepath,0,0);
 }


 return(err);
}



static pagesectionstr createpagesection;


/* tagged to UCREATES */
/* this function is always called last in undo/redo */

static os_error * uncreates(filestr * file) 
{
 frflflush(file);
// cprintf("vxflush");
 
 vxflush(file,VRBACK,createpagesection.page,createpagesection.section);
// cprintf("frflvalid");
 frflvalid(file);
// cprintf("frflscrap");
 frflscrap(file);
 vxscrap();

 return(undowritefn(file,UCREATEF,&createpagesection,sizeof(pagesectionstr)));
}


/* tagged to UCREATEF */

static os_error * uncreatef(filestr * file)
{
 undoread(file,&createpagesection,sizeof(pagesectionstr));

 vxstart(file);
 frflstart(file);

 return(undowritefn(file,UCREATES,NULL,0));
}






static os_error * framerepaint(frametoolstr * ftool,posnstr * posn,int all)
{
 os_error * err;
 aboxstr    abox;
 int        dx;
 int        dy;

 err=NULL;

 if(ftool->framecurtool==FRAME_RECT)
 {
  abox.x0=ftool->framestart.x;
  abox.x1=ftool->framelast.x;
  abox.y1=ftool->framestart.y;
  abox.y0=ftool->framelast.y;
 }
 else
 if(ftool->framecurtool==FRAME_ELLIPSE)
 {
  dx=ABS(ftool->framestart.x-ftool->framelast.x);
  dy=ABS(ftool->framestart.y-ftool->framelast.y);

  abox.x0=ftool->framestart.x-dx;
  abox.x1=ftool->framestart.x+dx;
  abox.y1=ftool->framestart.y-dy;
  abox.y0=ftool->framestart.y+dy;
 }
 else
 if(ftool->framecurtool==FRAME_STRSEG || ftool->framecurtool==FRAME_CURSEG)
 {
  abox.x0=ftool->framestart.xps-ftool->framestart.xpo;
  abox.x1=ftool->framestart.xpe-ftool->framestart.xpo;
  abox.y1=ftool->framestart.ype-ftool->framestart.ypo;
  abox.y0=ftool->framestart.yps-ftool->framestart.ypo;

  if(!all)
  {
   if(ftool->framecurtool==FRAME_CURSEG ||
      ftool->framepath[ftool->framepathptr1]==DRAWCURVE) 
                                         ftool->framepathcatch=2;
   else                                  ftool->framepathcatch=1;
  }
 }

 err=viewrefresharea(posn->file,posn->page,&abox,VRTOOLS);

 ftool->framepathcatch=0;

 return(err);
}



static os_error * framereset(filestr * file,viewstr * view)
{
 toolrecordstr * record;

 toolrecord(file,&record);

 if(!record->adjust && record->prevtoolkey!=record->toolkey)
                     return(toolselect(TL_SELECT,0,view,file));
 else
                     return(NULL);
}



#define ACTMINW 6000
#define ACTMINH 6000


/* click that creates an object */

static os_error * framecreate(frametoolstr * ftool,filestr * file)
{
 os_error * err;
 objectloc  loc;
 pagestr  * page;
 curvestr   c;
 int        curveok;

 err=NULL;


 err=framerepaint(ftool,&ftool->framestart,1);
 ftool->framestate=0;
 iprefreshpane(file,0x1,IP_PANE1);
 rulermarkend();


 curveok=1;

 if(curveok)
 {
  if(ftool->frametool==FRAME_RECT)
  {
   newcurve(&c,CBOX,NULL,0,file);
   c.box.x=MIN(ftool->framestart.x,ftool->framelast.x);
   c.box.y=MIN(ftool->framestart.y,ftool->framelast.y);
   c.box.w=ABS(ftool->framestart.x-ftool->framelast.x);
   c.box.h=ABS(ftool->framestart.y-ftool->framelast.y);
  }
  else
  if(ftool->frametool==FRAME_ELLIPSE)
  {
   newcurve(&c,CELLIPSE,NULL,0,file);
   c.box.w=2*ABS(ftool->framestart.x-ftool->framelast.x);
   c.box.h=2*ABS(ftool->framestart.y-ftool->framelast.y);
   c.box.x=ftool->framestart.x-c.box.w/2;
   c.box.y=ftool->framestart.y-c.box.h/2;
  }
  else
  if(ftool->frametool==FRAME_STRSEG || ftool->frametool==FRAME_CURSEG)
  {
   /* create an object, and put the path in it */

   /* put a cut here for null curves */
   /* frames with less than 3 sides  */


   if((ftool->framepathptr1==ftool->framepathptr2) || 
      (ftool->framepathptr2==ftool->framepathptr3 && 
                         ftool->framemode==FRAME_FRAME))curveok=0;
   else
   {
    err=frameclose(ftool,ftool->framemode==FRAME_FRAME);

    newcurve(&c,CPATH,&ftool->framepath,
                     ftool->framepathptr*sizeof(int),file);

   /* getpathbox(&c.box,ftool->framepath); */

    validpath(&c.box,&c.data.path.flags,ftool->framepath);
   }
  }
 }

 curveok=curveok && ( ((c.box.w>ACTMINW) && (c.box.h>ACTMINH)) || 
                      (ftool->framemode==FRAME_DIAGRAM && 
                                 (c.box.w>ACTMINW || c.box.h>ACTMINH)) );

 if(!err && curveok)
 {
  loc.file=file;
  loc.section=ftool->framestart.section;
  loc.page=ftool->framestart.page;
  loc.parent=loc.page;

  page=(pagestr*)fob_find(loc.page,loc.file);
  loc.after=endofchain(page->chain.child,loc.file);
  loc.chain=0;

  err=modfile(loc.file,loc.section);
  if(!err)
  {
   undostart(loc.file);

   undowritefn(loc.file,UCREATES,NULL,0);
   vxstart(loc.file);
   frflstart(loc.file);

   if(ftool->framemode==FRAME_FRAME)   err=objectcreate(&loc,OTEXTBOX,&c);
   else
   if(ftool->framemode==FRAME_DIAGRAM) err=objectcreate(&loc,OLINE,&c);

   createpagesection.page=loc.page;
   createpagesection.section=loc.section;
   uncreates(loc.file);

//   cprintf("create frame");

   clearselectionlo();
   objectselect((objhdrstr *)fob_find(loc.object,loc.file),&ftool->framestart,SELECTOBJECT);

   framereset(loc.file,ftool->framestart.view);

   undoend(loc.file);
  }

  deletecurve(&c,file);
 }

 return(err);
}



static void framelastcur(frametoolstr * ftool,posnstr * posn)
{
 int * framepath;
 int   framepathptr;

 framepath=ftool->framepath;
 framepathptr=ftool->framepathptr;

 framepath[framepathptr]=DRAWCURVE;
 framepath[framepathptr+1]=posn->x;
 framepath[framepathptr+2]=posn->y;
 framepath[framepathptr+3]=posn->x;
 framepath[framepathptr+4]=posn->y;
 framepath[framepathptr+5]=posn->x;
 framepath[framepathptr+6]=posn->y;
 framepath[framepathptr+7]=DRAWEND;
}



static void framelaststr(frametoolstr * ftool,posnstr * posn)
{
 int * framepath;
 int   framepathptr;

 framepath=ftool->framepath;
 framepathptr=ftool->framepathptr;

 framepath[framepathptr]=DRAWDRAW;
 framepath[framepathptr+1]=posn->x;
 framepath[framepathptr+2]=posn->y;
 framepath[framepathptr+3]=DRAWEND;
}



static os_error * framecancelsub(frametoolstr * ftool)
{
 os_error * err;

 err=NULL;

 if(ftool->framestate)
 {
  err=framerepaint(ftool,&ftool->framestart,1);
  ftool->framestate=0;
  framedrag=0;
//  cprintf("toolsstopdrag");

  toolsstopdrag();
  iprefreshpane(ftool->framestart.file,0x1,IP_PANE1);
 }

 flex_free((flex_ptr)&ftool->framepath);

 rulermarkend();

 return(err);
}



/* polygon tool stuff */

static os_error * frameaddpoint(frametoolstr * ftool,posnstr * posn)
{
 os_error * err;

 err=NULL;

 if(posn->valid)
 {
  err=flex_chunk((flex_ptr)&ftool->framepath,
                 (ftool->framepathptr+8+7)*sizeof(int),FRAMECHUNK);
               /* (8+7) is (4+3) for straight bits but may change to curve */

  if(ftool->frametool==FRAME_STRSEG)
  {
   if(ftool->framepath[ftool->framepathptr1]==DRAWCURVE)
   {
    ftool->framepathcatch=2;
    err=viewrefreshall(posn->file,posn->view->branch,VRTOOLS);
    ftool->framepathcatch=3;
   }
   else
   {
    ftool->framepathcatch=4;
   }

   ftool->framepath[ftool->framepathptr+1]=posn->x;
   ftool->framepath[ftool->framepathptr+2]=posn->y;

   ftool->framepathptr4=ftool->framepathptr3;
   ftool->framepathptr3=ftool->framepathptr2;
   ftool->framepathptr2=ftool->framepathptr1;
   ftool->framepathptr1=ftool->framepathptr;
   ftool->framepathptr+=3;

   framelaststr(ftool,posn);
  }
  else
  if(ftool->frametool==FRAME_CURSEG)
  {
   ftool->framepathcatch=2;
   err=viewrefreshall(posn->file,posn->view->branch,VRTOOLS);

   ftool->framepath[ftool->framepathptr+5]=posn->x;
   ftool->framepath[ftool->framepathptr+6]=posn->y;

   ftool->framepathptr4=ftool->framepathptr3;
   ftool->framepathptr3=ftool->framepathptr2;
   ftool->framepathptr2=ftool->framepathptr1;
   ftool->framepathptr1=ftool->framepathptr;
   ftool->framepathptr+=7;

   framelastcur(ftool,posn);

   ftool->framepathcatch=3;
  }

  ftool->seg.x=posn->x;
  ftool->seg.y=posn->y;

  calctangents(ftool->framepath,0,0);    /* the lot */
  err=viewrefreshall(posn->file,posn->view->branch,VRTOOLS);
  ftool->framepathcatch=0;
 }

 return(err);
}




void framesnap(posnstr * posn)
{
 viewstr * view;
 int       x;
 int       y;

 view=posn->view;

 x=posn->x;
 y=posn->y;

 if(!(view->data.grid.lock & GRIDLOCK))
 {
  posn->object=0;
  snap1(&x,&y,posn);
 }

 if(view->data.grid.lock)
 {
  if(view->data.grid.lock & GRIDLOCK) x=gridsnapx(posn->x,view,0);
  y=gridsnapy(posn->y,view,0);
 }

 posn->x=x;
 posn->y=y;
}




static os_error * framebegin(frametoolstr * ftool,posnstr * posn)
{
 os_error * err;
 rmstr      rmarks;

 err=NULL;

 ftool->framestart=*posn;
 framesnap(&ftool->framestart);

 ftool->framelast=ftool->framestart;
 ftool->framestate=1;
 ftool->seg.x=ftool->framestart.x;
 ftool->seg.y=ftool->framestart.y;

 if(ftool->frametool==FRAME_CURSEG || ftool->frametool==FRAME_STRSEG)
 {
  if(ftool->framepath) 
             err=flex_chunk((flex_ptr)&ftool->framepath,FRAMECHUNK,FRAMECHUNK);
  else       err=flex_alloc((flex_ptr)&ftool->framepath,FRAMECHUNK);

  ftool->framepathptr=0;

  if(!err)
  {
   ftool->framepathptr4=0;
   ftool->framepathptr3=0;
   ftool->framepathptr2=0;
   ftool->framepathptr1=0;
   ftool->framepath[ftool->framepathptr1]=DRAWMOVE;
   ftool->framepath[ftool->framepathptr1+1]=ftool->framestart.x;
   ftool->framepath[ftool->framepathptr1+2]=ftool->framestart.y;
   ftool->framepathptr=3;

   if(ftool->frametool==FRAME_STRSEG)
   {
    framelaststr(ftool,posn);
   }
   else
   if(ftool->frametool==FRAME_CURSEG)
   {
    framelastcur(ftool,posn);
    calctangents(ftool->framepath+ftool->framepathptr3,
                               (ftool->framepathptr3>0)+
                               (ftool->framepathptr2>ftool->framepathptr3),0);
   }

   framedrag=2;
   toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGPTR);
  }
 }
 else
 {
  rmarks.ay[0]=rmarks.ay[1]=1;
  rmarks.y[0]=rmarks.y[1]=ftool->framestart.y;
  rmarks.ax[0]=rmarks.ax[1]=1;
  rmarks.x[0]=rmarks.x[1]=ftool->framestart.x;

  toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGPTR);
 }

 err=viewrefreshall(posn->file,posn->view->branch,VRTOOLS);
 iprefreshpane(posn->file,0x1,IP_PANE1);
 rulermarkstart(&rmarks,posn->page,posn->branch,posn->file);

 return(err);
}


static viewstr * nosnapview;


static os_error * framekey(toolboxstr * toolbox,int * key,fkeyfn * fk)
{
 menukey(key);
 if(*key==-1) *fk=menukey;
 else
 {
  toolnudgekey(key);  /* don't claim next key press or else       */
                      /* will end up with key presses permanently */
  if(*key==-1) nosnapview=menuview;
 }

 return(NULL);
 USE(toolbox);
}




static os_error * framedragpoint(frametoolstr * ftool,posnstr * posn)
{
 if(ftool->framecurtool==FRAME_CURSEG) ftool->framepathcatch=3;
 else                                  ftool->framepathcatch=2;

 framerepaint(ftool,posn,1);

 ftool->framepath[ftool->framepathptr]=DRAWEND;

 ftool->framepathptr=ftool->framepathptr1;
 ftool->framepathptr1=ftool->framepathptr2;
 ftool->framepathptr2=ftool->framepathptr3;
 ftool->framepathptr3=ftool->framepathptr4;

 if(ftool->framepath[ftool->framepathptr1]==DRAWCURVE)
 {
  ftool->seg.x=ftool->framepath[ftool->framepathptr1+5];
  ftool->seg.y=ftool->framepath[ftool->framepathptr1+6];
 }
 else
 {
  ftool->seg.x=ftool->framepath[ftool->framepathptr1+1];
  ftool->seg.y=ftool->framepath[ftool->framepathptr1+2];
 }

 framedrag=1;
 framerepaint(ftool,posn,0);
 toolsstartdrag(posn,TOOLDRAGPAGE|TOOLDRAGPTR);

 return(NULL);

 USE(ftool);
}



static os_error * frameclick(toolboxstr * toolbox,posnstr * posn)
{
 os_error     * err;
 frametoolstr * ftool;
 int            buttonm;

 err=NULL;


 if(!posn->ghost)
 {
  ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

  buttonm=posn->buttons & (~VBSHIFT);

  if(buttonm==VBCLICKS || 
     buttonm==VBLONGCLICKS ||
     buttonm==VBDRAGS  || 
     buttonm==VBCLICKA ||
     buttonm==VBDBLCLICKS)
  {
   /* It would be better not to start a drag with a click          */
   /* Leads to oddities, like create frame, rapidly move mouse out */
   /* of window and click menu.                                    */

   if(ftool->framestate)  /* we were drawing a box */
   {
    if((buttonm==VBDRAGS || buttonm==VBLONGCLICKS) && 
       (ftool->frametool==FRAME_STRSEG || ftool->frametool==FRAME_CURSEG))
    {
     err=framedragpoint(ftool,posn);
    }
    else
    if(buttonm==VBCLICKS &&
       (ftool->frametool==FRAME_STRSEG || ftool->frametool==FRAME_CURSEG))
    {
     frameaddpoint(ftool,&ftool->framelast);
    }
    else
    if(buttonm!=VBDRAGS && buttonm!=VBLONGCLICKS)
    {
     err=framecreate(ftool,posn->file);
     framedrag=0;
    }
   }
   else  /* not in frame state */
   {
    if((buttonm==VBCLICKS && 
        (ftool->frametool==FRAME_STRSEG || ftool->frametool==FRAME_CURSEG)) ||
       buttonm==VBDRAGS ||
       buttonm==VBLONGCLICKS)
    {
     err=framebegin(ftool,posn);
    }
   }
  }
 }

 caretensure(posn->view);

 return(err);
}




static os_error * framemove(toolboxstr * toolbox,posnstr * xposn,int move)
{
 frametoolstr * ftool;
 int            dx;
 int            dy;
 int            ax;
 int            ay;
 int            sx;
 int            sy;
 posnstr        posn;
 rmstr          rmarks;


 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 toolschangepointer(0,USERPOINTER,&drawptr);

 if(move && ftool->framestate && !xposn->ghost && 
                                  xposn->page==ftool->framestart.page)
 {
  posn=*xposn;

  if(posn.view!=nosnapview) framesnap(&posn);

  framerepaint(ftool,&posn,0);

  if(ftool->frametool==FRAME_STRSEG || ftool->frametool==FRAME_CURSEG)
  {
   if(isshift)
   {
    dx=posn.x-ftool->seg.x;
    dy=posn.y-ftool->seg.y;

    ax=ABS(dx);
    ay=ABS(dy);

    if(ax>ay) posn.y=ftool->seg.y;
    else      posn.x=ftool->seg.x;
   }

   ftool->framelast=posn;

   if(ftool->frametool==FRAME_STRSEG)
   {
    ftool->framepath[ftool->framepathptr+1]=posn.x;
    ftool->framepath[ftool->framepathptr+2]=posn.y;
   }
   else    /* ftool->frametool==FRAME_CURSEG */
   {
    ftool->framepath[ftool->framepathptr+5]=posn.x;
    ftool->framepath[ftool->framepathptr+6]=posn.y;
   }

   calctangents(ftool->framepath+ftool->framepathptr3,
               (ftool->framepathptr3>0)+
               (ftool->framepathptr2>ftool->framepathptr3),0);
  }
  else
  {
   if(isshift)
   {
    dx=posn.x-ftool->framestart.x;
    dy=posn.y-ftool->framestart.y;

    ax=ABS(dx);
    ay=ABS(dy);
    sx=SGN(dx);
    sy=SGN(dy);

    dx=sx*MAX(ax,ay);
    dy=sy*MAX(ax,ay);
  
    ftool->framelast.x=ftool->framestart.x+dx;
    ftool->framelast.y=ftool->framestart.y+dy;
   }
   else
   {
    ftool->framelast=posn;
   }
  }

  rmarks.ay[0]=rmarks.ay[1]=1;
  rmarks.y[0]=ftool->framestart.y;
  rmarks.y[1]=ftool->framelast.y;

  rmarks.ax[0]=rmarks.ax[1]=1;
  rmarks.x[0]=ftool->framestart.x;
  rmarks.x[1]=ftool->framelast.x;

  rulermarkupdate(&rmarks);

  framerepaint(ftool,&posn,0);
  iprefreshpane(posn.file,0x1,IP_PANE1);
 }

 nosnapview=NULL;

 return(NULL);
}



/* maybe good idea to call framemove() to update to posn          */
/* it is not a good idea to use the passed posn, may not be valid */

static os_error * framedragend(toolboxstr * toolbox,
                               posnstr * posn,wimp_box * box)
{
 os_error     * err;
 frametoolstr * ftool;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 if(framedrag==1)
 {
  framerepaint(ftool,posn,0);
  framedrag=0;
  framerepaint(ftool,posn,0);
  err=frameaddpoint(ftool,&ftool->framelast);
 }
 else
 if(framedrag==2)
 {
  err=NULL;
 }
 else
 {
  err=framecreate(ftool,posn->file);
 }

 return(err);

 USE(box);
}












#define FIPX 0
#define FIPY 1
#define FIPH 3
#define FIPW 2


static iconmap fipmap[]=
{
 9,0,0,0,
 3,0,0,0,
 5,0,0,0,
 7,0,0,0,


 0,0,0,0,      /* terminator */
};






static os_error * fiprefresh(viewstr * view,int handle,int code)
{
 filestr      * file;
 toolboxstr   * toolbox;
 frametoolstr * ftool;
 int            mag;

 file=view->file;
 toolbox=file->toolbox;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 if(code==-1 || code==0x7FFFFFFF)
 {
  writeiconf(view->info,IP_HELPI,"%s",
   ftool->framemode==FRAME_FRAME?"{FRM00}":"{FRM01}");
 }

 if(code & 0x1)
 {
  if(ftool->framestate)
  {
   writeiconc(handle,fipmap[FIPX].realicon,
      unitstostring(ftool->framestart.x-prefp(file)->rulerx,view->data.units));
   writeiconc(handle,fipmap[FIPY].realicon,
      unitstostring(ftool->framestart.y-prefp(file)->rulery,view->data.units));

   mag=(ftool->framecurtool==FRAME_ELLIPSE)?2:1;

   writeiconc(handle,fipmap[FIPW].realicon,
       unitstostring((ftool->framelast.x-ftool->framestart.x)*mag,
                                                     view->data.units));

   writeiconc(handle,fipmap[FIPH].realicon,
       unitstostring((ftool->framelast.y-ftool->framestart.y)*mag,
                                                     view->data.units));
  }
  else
  {
   writeiconc(handle,fipmap[FIPX].realicon,"");
   writeiconc(handle,fipmap[FIPY].realicon,"");
   writeiconc(handle,fipmap[FIPW].realicon,"");
   writeiconc(handle,fipmap[FIPH].realicon,"");
  }
 }

 return(NULL);
}




static ipstr fipstr=
{
 NULL,         /* click    */
 NULL,         /* key      */
 fipmap,       /* icon map */
 TSELECTTOOL,  /* template */
 NULL,         /* redraw   */
 fiprefresh,   /* refresh  */
};




static os_error * frameselects(toolboxstr * toolbox,viewstr * view,
                                                    int select,int mode)
{
 frametoolstr * ftool;
 filestr      * file;

 file=view->file;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 if(select)
 {
  if(ftool->framemode!=mode)
  {
   SWOP(ftool->framecurtool,ftool->framediagramtool);
   ftool->framemode=mode;
   if(ftool->frametool!=ftool->framecurtool)
                             toolselect(ftool->framecurtool,0,view,file);
  }
 }
 else
 {
  framecancelsub(ftool);
 }

 if(!undoactive(file))
 {
  if(select) clearselectionlo();
 }

 ipchangepane(toolbox,IP_PANE1,select?&fipstr:NULL);

 return(NULL);
}


static os_error * frameselect(toolboxstr * toolbox,viewstr * view,
                                                   filestr * file,int select)
{
 return(frameselects(toolbox,view,select,FRAME_FRAME));
 USE(file);
}


static os_error * diagramselect(toolboxstr * toolbox,viewstr * view,
                                                     filestr * file,int select)
{
 return(frameselects(toolbox,view,select,FRAME_DIAGRAM));
 USE(file);
}




static os_error * frameptr(toolboxstr * toolbox,posnstr * posn)
{
 toolschangepointer(0,USERPOINTER,&drawptr);

 return(NULL);

 USE(toolbox);
 USE(posn);
}



static os_error * framecancel(toolboxstr * toolbox,viewstr * view,
                                                   filestr * file)
{
 frametoolstr * ftool;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 return(framecancelsub(ftool));

 USE(view);
 USE(file);
}



static os_error * frameinit(toolboxstr * toolbox,int code)
{
 frametoolstr * ftool;

 if(toolbox)
 {
  ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

  if(code==TINITCREATE)
  {
   ftool->framemode=FRAME_FRAME;
   ftool->frametool=ftool->framecurtool=FRAME_RECT;
   ftool->framediagramtool=FRAME_STRSEG;
  }
  else
  if(code==TINITDELETE)
  {
   flex_free((flex_ptr)&ftool->framepath);
  }
 }

 return(NULL);
}



static os_error * fmess(toolboxstr * toolbox,filestr * file,
                                      int message,void * data)
{

 if(message==TSAVEPOSN) ssaveselection(file);

 return(NULL);

 USE(toolbox);
 USE(data);
}



static os_error * smenu(toolboxstr * toolbox,viewstr * view,
                                             filestr * file,wimp_mousestr * m)
{
 return(openmenu("main_menu"));
 USE(view);
 USE(file);
 USE(m);
 USE(toolbox);
}




static toolstr frame=
{
 TL_FRAME,
 "frame",
 "t_frame\0t_pframe",
 "{FRATIP}",
 1,
 frameclick,
 framemove,
 frameselect,
 frameptr,
 framecancel,
 frameredraw,
 framedragend,
 smenu,
 framekey,
 NULL,
 frameinit,
 fmess,
 0,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */


 TPOINTER,
 0,
 sizeof(frametoolstr),
 1,       /*  rank      */
};


static toolstr diagram=
{
 TL_DIAGRAM,
 "line",
 "t_line\0t_pline",
 "{LINTIP}",
 1,
 frameclick,
 framemove,
 diagramselect,
 frameptr,
 framecancel,
 frameredraw,
 framedragend,
 smenu,
 framekey,
 NULL,
 NULL,
 fmess,
 0,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TPOINTER,
 0,
 0,
 2,       /*  rank      */
};


/*****************************************************************************/


static os_error * rectangleselect(toolboxstr * toolbox,viewstr * view,
                                                filestr * file,int select)
{
 frametoolstr * ftool;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 framecancelsub(ftool);

 if(select) ftool->frametool=ftool->framecurtool=FRAME_RECT;
 else       ftool->frametool=FRAME_NONE;

 return(NULL);

 USE(view);
 USE(file);
}


static os_error * ellipseselect(toolboxstr * toolbox,viewstr * view,
                                              filestr * file,int select)
{
 frametoolstr * ftool;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 framecancelsub(ftool);

 if(select) ftool->frametool=ftool->framecurtool=FRAME_ELLIPSE;
 else       ftool->frametool=FRAME_NONE;

 return(NULL);

 USE(view);
 USE(file);
 USE(toolbox);
}


static os_error * curselect(toolboxstr * toolbox,viewstr * view,
                                          filestr * file,int select)
{
 frametoolstr * ftool;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 if(select)
 {
  ftool->frametool=ftool->framecurtool=FRAME_CURSEG;

  if(ftool->framestate)
  {
   framerepaint(ftool,&ftool->framelast,0);
   framelastcur(ftool,&ftool->framelast);
   calctangents(ftool->framepath+ftool->framepathptr3,
               (ftool->framepathptr3>0)+
               (ftool->framepathptr2>ftool->framepathptr3),0);

   framerepaint(ftool,&ftool->framelast,0);
  }
 }
 else       ftool->frametool=FRAME_NONE;

 return(NULL);

 USE(view);
 USE(file);
 USE(toolbox);
}


static os_error * strselect(toolboxstr * toolbox,viewstr * view,
                                          filestr * file,int select)
{
 frametoolstr * ftool;

 ftool=(frametoolstr*)toolbox->toolstate[toolindex(TL_FRAME)].hprivate;

 if(select)
 {
  ftool->frametool=ftool->framecurtool=FRAME_STRSEG;

  if(ftool->framestate)
  {
   framerepaint(ftool,&ftool->framelast,0);
   framelaststr(ftool,&ftool->framelast);
   calctangents(ftool->framepath+ftool->framepathptr3,
               (ftool->framepathptr3>0)+
               (ftool->framepathptr2>ftool->framepathptr3),0);

   framerepaint(ftool,&ftool->framelast,0);
  }
 }
 else ftool->frametool=FRAME_NONE;

 return(NULL);

 USE(view);
 USE(file);
 USE(toolbox);
}



static toolstr rectangle=
{
 TL_RECTANGLE,
 "rect",
 "t_rect\0t_prect",
 "{RECTIP}",
 1,
 NULL,
 NULL,
 rectangleselect,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,

 1,

 1,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TRADIO,
 0,
 0,
 0x100,   /*  rank      */
};

static toolstr ellipse=
{
 TL_ELLIPSE,
 "ellipse",
 "t_ellipse\0t_pellipse",
 "{ELITIP}",
 1,
 NULL,
 NULL,
 ellipseselect,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,

 1,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TRADIO,
 0,
 0,
 0x101,   /*  rank      */
};

static toolstr curseg=
{
 TL_CURSEG,
 "curseg",
 "t_curseg\0t_pcurseg",
 "{CRVTIP}",
 1,
 NULL,
 NULL,
 curselect,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,

 1,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TRADIO,
 0,
 0,
 0x103,   /*  rank      */
};

static toolstr strseg=
{
 TL_STRSEG,
 "strseg",
 "t_strseg\0t_pstrseg",
 "{STRTIP}",
 1,
 NULL,
 NULL,
 strselect,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,
 NULL,

 1,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TRADIO,
 0,
 0,
 0x102,   /* rank       */
};


/*****************************************************************************/


#define UNDOS 2

static undofnstr undos[UNDOS]=
{
 uncreates,       NULL, UCREATES,      "{UCREATE}",
 uncreatef,       NULL, UCREATEF,      "{UCREATE}",
};




os_error * frametoolinit(void)
{
 os_error * err;

          err=toolregister(&frame);
 if(!err) err=toolregister(&diagram);
 if(!err) err=toolregister(&rectangle);
 if(!err) err=toolregister(&ellipse);
 if(!err) err=toolregister(&strseg);
 if(!err) err=toolregister(&curseg);

 if(!err) err=undoaddblock(undos,UNDOS);

 return(err);
}

