/*->c.xredraw */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"
#include "xprocess.h"
#include "BezierArc.h"

#include "DrawMod.h"
#include "deb.h"
#include "err.h"
#include "xmath.h"
#include "colour.h"
#include "etc.h"

#include "str.h"

#include "view.h"
#include "geometry.h"
#include "xrender.h"

#include "xredraw.h"


static int actstash;
static int colourstash;
static int stylestash;
static int validwcolour;
static int validgcolour;
static int validstyle;
static int tomask;


#define MITRELIMIT 0x80000





os_error * rbbc_rectanglefill(boxstr * box,viewredrawstr * v)
{
 int x0;
 int x1;
 int y0;
 int y1;

 x0=v->ox+osunits(v->xpo+box->x,v->view);
 x1=v->ox+osunits(v->xpo+box->x+box->w,v->view);
 y0=v->oy-osunits(v->ypo+box->y,v->view);
 y1=v->oy-osunits(v->ypo+box->y+box->h,v->view);

 x0=(x0+cvdu.deltax-1)&cvdu.maskx;
 x1=(x1+cvdu.deltax-1)&cvdu.maskx;
 y0=(y0+cvdu.deltay-1)&cvdu.masky;
 y1=(y1+cvdu.deltay-1)&cvdu.masky;

 return(bbc_rectanglefill(x0,y1,x1-x0,y0-y1));
}





os_error * rbbc_rectangle(boxstr * box,viewredrawstr * v)
{
 int x0;
 int x1;
 int y0;
 int y1;

 x0=v->ox+osunits(v->xpo+box->x,v->view);
 x1=v->ox+osunits(v->xpo+box->x+box->w,v->view);
 y0=v->oy-osunits(v->ypo+box->y,v->view);
 y1=v->oy-osunits(v->ypo+box->y+box->h,v->view);

 x0=(x0+cvdu.deltax-1)&cvdu.maskx;
 x1=(x1+cvdu.deltax-1)&cvdu.maskx;
 y0=(y0+cvdu.deltay-1)&cvdu.masky;
 y1=(y1+cvdu.deltay-1)&cvdu.masky;

 return(bbc_rectangle(x0,y1,x1-x0,y0-y1));
}








os_error * rbbc_move(int x,int y,viewredrawstr * v)
{
 x=v->ox+osunits(v->xpo+x,v->view);  
 y=v->oy-osunits(v->ypo+y,v->view);  
 return(bbc_move(x,y));
}


os_error * rbbc_draw(int x,int y,viewredrawstr * v)
{
 x=v->ox+osunits(v->xpo+x,v->view);  
 y=v->oy-osunits(v->ypo+y,v->view);  
 return(bbc_draw(x,y));
}



os_error * rtbbc_move(int x,int y,transformstr * tr,viewredrawstr * v)
{
 coordstr c;
 c.x=x;
 c.y=y;
 if(tr) tr_transformset(&c,1,tr);
 x=v->ox+osunits(v->xpo+c.x,v->view);  
 y=v->oy-osunits(v->ypo+c.y,v->view);  
 return(bbc_move(x,y));
}


os_error * rtbbc_draw(int x,int y,transformstr * tr,viewredrawstr * v)
{
 coordstr c;

// cprintf("rtbbc_draw");

 c.x=x;
 c.y=y;
 if(tr) tr_transformset(&c,1,tr);
 x=v->ox+osunits(v->xpo+c.x,v->view);  
 y=v->oy-osunits(v->ypo+c.y,v->view);  
 return(bbc_draw(x,y));
}


os_error * rtbbc_drawstyle(int x,int y,transformstr * tr,viewredrawstr * v)
{
 coordstr c;
 c.x=x;
 c.y=y;
 if(tr) tr_transformset(&c,1,tr);
 x=v->ox+osunits(v->xpo+c.x,v->view);  
 y=v->oy-osunits(v->ypo+c.y,v->view);  
 return(bbc_plot(bbc_DottedBoth+5,x,y));
}




/* we would like the default value of a dot pattern with a repeat of 8 */
/* fx163,242,8                                                         */


os_error * rsetdotstyle(int style)
{
 os_error * err;

 if(validstyle && style==stylestash) return(NULL);
// err=bbc_vduq(23,6,style);
 err=NULL;
 validstyle=1;
 stylestash=style;

 return(err);
}


os_error * rwimp_setcolour(int colour)
{
 if(tomask || (validwcolour && colourstash==colour)) return(NULL);
 validwcolour=1;
 validgcolour=0;
 colourstash=colour;
 return(wimp_setcolour(colour));
}


/* given a palette entry, return the closest gcol and set it up */

os_error * rsetgcol(int pal,int action)
{
 if(tomask || (validgcolour && colourstash==pal && action==actstash))
                                                                 return(NULL);
 validgcolour=1;
 validwcolour=0;
 colourstash=pal;
 actstash=action;

 return(setgcolaction(pal,action));
}


os_error * rseteorcol(int back,int front)
{
 if(tomask) return(NULL);

 validgcolour=0;
 validwcolour=0;
 return(seteorcol(back,front));
}



os_error * rseteorcolpal(int back,int front)
{
 if(tomask) return(NULL);

 validgcolour=0;
 validwcolour=0;
 return(seteorcolpal(back,front));
}



os_error * startredraw(void)
{
 validwcolour=0;
 validgcolour=0;
 validstyle=0;
 return(NULL);
}


void redrawswitch(int mask)
{
 tomask=mask;

// if(mask) setgcolour(cvdu.ncolours==63?255:cvdu.ncolours,0);
}




/* PLOT commands PRM page 286 or 1-607 */

/* in these functions if style>0  we have a dotted line */
/*                    if style==0 a line                */
/*                    if style<0  a fill                */



os_error * rrectangle(coordstr * point,int style,viewredrawstr * v)
{
 os_error * err;
 int        x0;
 int        y0;
 int        x1;
 int        y1;
 int        gx0;
 int        gx1;
 int        gy0;
 int        gy1;

 x0=v->ox+osunits(v->xpo+point[0].x,v->view);
 y0=v->oy-osunits(v->ypo+point[1].y,v->view);
 x1=v->ox+osunits(v->xpo+point[1].x,v->view);
 y1=v->oy-osunits(v->ypo+point[0].y,v->view);

 gx0=v->redrawstr.g.x0;
 gx1=v->redrawstr.g.x1;
 gy0=v->redrawstr.g.y0;
 gy1=v->redrawstr.g.y1;

 if(!(x0<gx0 && x1>=gx1 && y0<gy0 && y1>=gy1))
 {

  err=bbc_move(x0,y1);
  if(!err)
  {
   if(style>0)
   {
    err=rsetdotstyle(style);
    if(!err) err=bbc_plot(bbc_DottedBoth+5,x0,y0);
    if(!err) err=bbc_plot(bbc_DottedBoth+5,x1,y0);
    if(!err) err=bbc_plot(bbc_DottedBoth+5,x1,y1);
    if(!err) err=bbc_plot(bbc_DottedBoth+5,x0,y1);
   }
   else
   if(style==0)
   {
             err=bbc_plot(bbc_SolidExFinal+5,x0,y0);
    if(!err) err=bbc_plot(bbc_SolidExFinal+5,x1,y0);
    if(!err) err=bbc_plot(bbc_SolidExFinal+5,x1,y1);
    if(!err) err=bbc_plot(bbc_SolidExFinal+5,x0,y1);

   }
   else
   {
    err=bbc_plot(bbc_RectangleFill+5,x1,y0);
   }
  }
 }
 else err=NULL;

 return(err);
}




os_error * rparallel(coordstr * point,int style,viewredrawstr * v)
{
 os_error * err;
 coordstr   oscoord[4];
 viewstr  * view;
 int        i;

 view=v->view;

 for(i=0;i<3;i++)
 {
  oscoord[i].x=v->ox+osunits(v->xpo+point[i].x,view);
  oscoord[i].y=v->oy-osunits(v->ypo+point[i].y,view);
 }

 if(style>=0)
 {
  oscoord[3].x=oscoord[2].x-(oscoord[1].x-oscoord[0].x);
  oscoord[3].y=oscoord[2].y-(oscoord[1].y-oscoord[0].y);
 }

 err=bbc_move(oscoord[0].x,oscoord[0].y);
 if(!err)
 {
  if(style>0)
  {
   if(!err) err=rsetdotstyle(style);
   for(i=3;i>=0;i--) bbc_plot(bbc_DottedBoth+5,oscoord[i].x,oscoord[i].y);
  }
  else
  if(style==0)
  {
   for(i=3;i>=0;i--) bbc_plot(bbc_SolidExFinal+5,oscoord[i].x,oscoord[i].y);
  }
  else
  {
   err=bbc_move(oscoord[1].x,oscoord[1].y);
   if(!err) err=bbc_plot(bbc_ParallelFill+5,oscoord[2].x,oscoord[2].y);
  }
 }
 return(err);
}








os_error * rcross(coordstr * point,int style,int n,viewredrawstr * v)
{
 os_error * err;
 coordstr   oscoord[2];
 viewstr  * view;
 int        i;

 err=NULL;

 view=v->view;

 for(i=0;i<n;i+=2)
 {
  oscoord[0].x=v->ox+osunits(v->xpo+point[i].x,view);
  oscoord[0].y=v->oy-osunits(v->ypo+point[i].y,view);
  oscoord[1].x=v->ox+osunits(v->xpo+point[i+1].x,view);
  oscoord[1].y=v->oy-osunits(v->ypo+point[i+1].y,view);

  err=bbc_move(oscoord[0].x,oscoord[0].y);
  if(!err)
  {
   if(style>0)
   {
    if(!err) err=rsetdotstyle(style);
    bbc_plot(bbc_DottedBoth+5,oscoord[1].x,oscoord[1].y);
   }
   else
   if(style==0)
   {
    bbc_plot(bbc_SolidExFinal+5,oscoord[1].x,oscoord[1].y);
   }
  }
 }

 return(err);
}








os_error * rcircle(coordstr * point,int style,viewredrawstr * v)
{
 os_error * err;
 coordstr   oscoord[2];
 viewstr  * view;
 int        i;

 view=v->view;

 for(i=0;i<2;i++)
 {
  oscoord[i].x=v->ox+osunits(v->xpo+point[i].x,view);
  oscoord[i].y=v->oy-osunits(v->ypo+point[i].y,view);
 }

 err=bbc_move(oscoord[0].x,oscoord[0].y);
 if(!err)
 {
  if(style>=0)
  {
   err=rsetdotstyle(style);
   if(!err) err=bbc_plot(bbc_Circle+5,oscoord[1].x,oscoord[1].y);
  }
  else
  {
   err=bbc_plot(bbc_CircleFill+5,oscoord[1].x,oscoord[1].y); 
  }
 }
 return(err);
}



int linestyle2draw(int * dash,int style,int scale)
{
 int i;
 int n;
 unsigned int mask;
 int dx;
 unsigned int st;
 int state;
 int newstate;
 int min;

 if(style==-1) return(0);   /* dash defn  */
 else
 {
  st=style;

  min=0;
  if((st>>16)==(st & 0xFFFF))
  {
   min=16;
   if((st>>24)==((st>>16) & 0xFF))
   {
    min=24;
    if((st>>28)==((st>>24) & 0xF))
    {
     min=28;
     if((st>>30)==((st>>28) & 0x3))
     {
      min=30;
     }
    }
   }
  }

  mask=0x80000000;
  state=0;
  dx=0;
  n=0;

  for(i=31;i>=min;i--)
  {
   newstate=(st & mask)!=0;

   if(state!=newstate && i!=31)
   {
    dash[2+n]=dx;
    dx=0;
    n++;
   }

   state=newstate;

   dx+=scale;
   mask>>=1;
  }

  if(dx)
  {
   dash[2+n]=dx;
   n++;
  }

  dash[0]=0;
  dash[1]=n;

		cprintf("dash count %d",n);

  return((int)dash);
 }
}




os_error * drawboxsize(coordstr * cx,int * path,pathstyle * style)
{
 os_error   * err;
 int          caps[4];
 int          dash[DASHSIZE];
 int          windingrule;
 int          styleword;

 err=NULL;

 caps[0]=style->join+((style->endcaps)<<8)+((style->startcaps)<<16);
 caps[1]=MITRELIMIT;
 caps[2]=((style->endwidth)>>8)+(((style->endheight & 0xFFFF00))<<8);
 caps[3]=((style->startwidth)>>8)+(((style->startheight & 0xFFFF00))<<8);

 windingrule=0x30+style->windingrule;    /* thicken */
 styleword=linestyle2draw(dash,style->style,style->stylescale);

 Draw_Bounds(path,style->thickness,((ISW9X)?DRAWMOD_OSUNITS:0),caps,styleword,windingrule,cx);

 return(err);
}




os_error * rdrawpath(int * path,int t,int style,int stylescale,int join)
{
 os_error * err;
 int        caps[4];
 int        dash[DASHSIZE];

 caps[0]=join;
 caps[1]=MITRELIMIT;
 caps[2]=0;
 caps[3]=0;

 style=linestyle2draw(dash,style,stylescale);

 err=Draw_Stroke(path,t,DRAWMOD_OSUNITS,caps,NULL,style,0);

 return(err);
}





os_error * rpathlines(int * path,pathstyle * style,viewredrawstr * v,
                                                            transformstr * tr)
{
 os_error   * err;
 int          caps[4];
 int          dash[DASHSIZE];
 int          effth;
 int          min;
 coordstr     cscale;
 int          tempx;
 int          tempy;
 int          styleword;
 int          flatness;

 err=NULL;


/* flatness   */
 tr_getscale(tr,&cscale);
 tempx=scale(180*200,0x10000,cscale.x*cvdu.xdpi);
 tempy=scale(180*200,0x10000,cscale.y*cvdu.ydpi);
 flatness=squareroot(tempx*tempy,0);


 if(style->action & 0x2)
 {
//  cprintf("Draw_Fill");
  err=render_draw_fill(path,tr,style->windingrule,flatness,DRAWMOD_OSUNITS,v);
 }
 else
 if(style->action & 0x1)
 {
  min=MIN(cvdu.deltax,cvdu.deltay);
  if(osunits(style->thickness,v->view)<min && !v->isprinter &&
                         !style->startcaps && !style->endcaps) 
  {
   effth=0;
  }
  else
  {
   effth=style->thickness;  // think this is in 72000'ths
  }

  caps[0]=style->join+((style->endcaps)<<8)+((style->startcaps)<<16);
  caps[1]=MITRELIMIT;
  caps[2]=((style->endwidth)>>8)+(((style->endheight & 0xFFFF00))<<8);
  caps[3]=((style->startwidth)>>8)+(((style->startheight & 0xFFFF00))<<8);

  styleword=linestyle2draw(dash,style->style,style->stylescale);

  err=render_draw_stroke(path,effth,DRAWMOD_OSUNITS,caps,tr,styleword,flatness,v); // flags was 0x80000038
 }
 return(err);
}




os_error * rpathline(int * path,pathstyle * style,viewredrawstr * v,transformstr * tr)
{
 transformstr restr;
 render_getdrtr(&restr,tr,v);
 return(rpathlines(path,style,v,&restr));
}




os_error * rpathchunk(coordstr * cx,viewredrawstr * v,transformstr * t)
{
 os_error * err;
 viewstr  * view;
 int        path[12];
 int        i;
 int      * p;

 view=v->view;

 if(t) tr_transformset(cx,4,t);
 for(i=0;i<4;i++)
 {
  cx[i].x=dosunits(v->ox+osunits(v->xpo+cx[i].x,view));
  cx[i].y=dosunits(v->oy-osunits(v->ypo+cx[i].y,view));
 }

 p=path;

 *p++=DRAWMOVE;
 *p++=cx[0].x;
 *p++=cx[0].y;
 *p++=DRAWCURVE;
 *p++=cx[1].x;
 *p++=cx[1].y;
 *p++=cx[2].x;
 *p++=cx[2].y;
 *p++=cx[3].x;
 *p++=cx[3].y;
 *p++=DRAWEND;

 err=Draw_Stroke(path,0,DRAWMOD_OSUNITS,NULL,NULL,0,0); // flags was 0x80000000

 return(err);
}




os_error * rsimpleline(int * p,viewredrawstr * v,transformstr * t)
{
 os_error * err;
 int        over;
 int        c;
 int        x;
 int        y;
 coordstr   cx[4];

 err=NULL;
 over=0;
 x=y=0;    /* compiler */

 while(!over)
 {
  c=*p++;

  switch(c)
  {
   case DRAWMOVE:
                 x=*p++;
                 y=*p++;
                 rtbbc_move(x,y,t,v);
                 break;

   case DRAWDRAW:
                 x=*p++;
                 y=*p++;
                 rtbbc_draw(x,y,t,v);
                 break;

   case DRAWCURVE:
                  cx[0].x=x;
                  cx[0].y=y;
                  cx[1].x=*p++;
                  cx[1].y=*p++;
                  cx[2].x=*p++;
                  cx[2].y=*p++;
                  x=cx[3].x=*p++;
                  y=cx[3].y=*p++;
                  rpathchunk(cx,v,t);
                  rtbbc_move(x,y,t,v);
                  break;

   case DRAWEND:
                 over=1;
                 break;
  }
 }
 return(err);
}




os_error * visiblebox(viewredrawstr * v,transformstr * tr,boxstr * box,int * visible)
{
 os_error   * err;
 boxstr       maxbox;
 coordstr     cx1;
 coordstr     cx2;
 transformstr drtr;

 getmaxbox(&maxbox,box,tr);

 render_getdrtr(&drtr,NULL,v);
 cx1.x=maxbox.x;
 cx1.y=maxbox.y;
 cx2.x=maxbox.x+maxbox.w;
 cx2.y=maxbox.y+maxbox.h;
 tr_transformset(&cx1,1,&drtr);
 tr_transformset(&cx2,1,&drtr);

 err=render_gwinrectdr(cx1.x,cx2.y,cx2.x,cx1.y,visible,v);

 return(err);
}











