/*->c.gline */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"
#include "xprocess.h"

#include "err.h"
#include "fsx.h"
#include "bf.h"
#include "task.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "colour.h"
#include "etc.h"


#include "constants.h"
#include "str.h"
#include "object.h"


#include "viewr.h"
#include "view.h"
#include "file.h"
#include "undo.h"
#include "chain.h"
#include "xredraw.h"
#include "curve.h"
#include "posn.h"

#include "select.h"
#include "units.h"
#include "dscript.h"
#include "pref.h"

#include "xcolour.h"
#include "geometry.h"


#include "editpath.h"

#include "objectdb.h"


#include "mpage.h"

#include "dbmore.h"

#include "gline.h"
#include "dbgline.h"



#define UGLTRAN     0x3100
#define UGLDELETE   0x3103
#define UGLRECREATE 0x3104
#define URDELETES   0x3105
#define URDELETEF   0x3106
#define URCREATES   0x3107
#define URCREATEF   0x3108
#define ULMXLOCAL   0x3109
#define ULMXMASTER  0x310A
#define UMAKEMASTER 0x310B
#define UGMOVE      0x310C
#define UUDM        0x310D
#define URDM        0x310E







static os_error * transline(glinestr * glinep,filestr * file,
                                      transformfn trf,int flags);




static int           ignoresnapmessage;





static os_error * redrawline(viewredrawstr * v,glinestr * line)
{
 os_error * err;
 int        guidecolour;

 err=NULL;

 if(line->key==v->object) v->ignore=0;

 if(!v->ignore)
 {
  if(boxinredraw(v,&line->box))
  {
   guidecolour=
        (v->isprinter || !v->view->data.guides)?0:GUIDECOLOUR|
                                  OUTLINEWIMPLINE|(GUIDELINESTYLE<<8);

   outlineline(v,&line->fcurve,guidecolour,NULL,&line->transform,0);
  }
 }
 return(err);
}



#define DEFOUTSET MMSIZE


static os_error * createline(objectloc * loc,curvestr * c)
{
 os_error   * err;
 int          gline;
 glinestr   * glinep;
 curvestr     fcurve;


 err=fob_create(sizeof(glinestr),loc->file,OGLINE,(fobjptr)&glinep);
 if(!err)
 {
  gline=glinep->key;

  tr_setidentity(&glinep->transform);

  copycurve(&fcurve,loc->file,c,loc->file);

  glinep=(glinestr*)fob_find(gline,loc->file);

  glinep->fcurve=fcurve;

  getmaxbox(&glinep->box,&glinep->fcurve.box,&glinep->transform);

                /*loc->after,loc->parent */
  insafter(gline,0,loc->page,PGLINECHAIN,loc->file);

  loc->object=gline;

  viewrefreshobject(loc->object,loc->section,loc->page,loc->file);

  undowritefn(loc->file,UGLDELETE,&loc->object,sizeof(int));
 }
 return(err);
}



/*****************************************************************************/

/* move an object - within the same file */

static transformstr xtransform;

static int xtrf(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct)
{
 tr_multiply(r,t,&xtransform);

 return(0);

 USE(c);
 USE(ct);
}



typedef struct gxmovestr
{
 int            gline;
 int            spage;
 int            dpage;
 int            x;
 int            y;
} gxmovestr;



static os_error * unlinemove(filestr * file)
{
 gxmovestr    old;
 gxmovestr    xnew;
 glinestr   * glinep;

 undoread(file,&xnew,sizeof(gxmovestr));

 old=xnew;
 old.spage=xnew.dpage;
 old.dpage=xnew.spage;

 glinep=(glinestr*)fob_find(old.gline,file);

 old.x=glinep->fcurve.box.x;
 old.y=glinep->fcurve.box.y;

 undowritefn(file,UGMOVE,&old,sizeof(gxmovestr));

 glinep=(glinestr*)fob_find(xnew.gline,file);
 glinep->fcurve.box.x=xnew.x;
 glinep->fcurve.box.y=xnew.y;

 vxaddn((sobjectstr**)&glinep,file,xnew.spage);

 unlinkobject(old.gline,PGLINECHAIN,file);
 insafter(xnew.gline,0,xnew.dpage,PGLINECHAIN,file);

 glinep=(glinestr*)fob_find(xnew.gline,file);

 getmaxbox(&glinep->box,&glinep->fcurve.box,&glinep->transform);

 vxaddn((sobjectstr **)&glinep,file,xnew.dpage);

 return(NULL);
}





static os_error * linecopym(objectloc * src,objectloc * dest,transformstr * tr)
{
 glinestr   * glinep;
 filestr    * sfile;
 int          object;
 int          trmode;
 gxmovestr    old;

 sfile=src->file;
 object=src->object;

 glinep=(glinestr*)fob_find(object,sfile);
 vxaddn((sobjectstr**)&glinep,sfile,src->page);

 if(dest->copymode & LOCUNDO)
 {
  old.gline=object;
  old.spage=dest->page;
  old.dpage=src->page;
  old.x=glinep->fcurve.box.x;
  old.y=glinep->fcurve.box.y;

  undowritefn(sfile,UGMOVE,&old,sizeof(gxmovestr));
 }

 unlinkobject(object,PGLINECHAIN,sfile);
 dest->object=object;

                 /*dest->after,dest->parent */
 insafter(object,0,dest->page,PGLINECHAIN,sfile);

 glinep=(glinestr*)fob_find(object,sfile);
 xtransform=*tr;

 trmode=TRNOUPDATE|TRNOREFLOWS|TRNOUNDO;

 transline(glinep,sfile,xtrf,trmode);

 glinep=(glinestr*)fob_find(object,sfile);
 vxaddn((sobjectstr**)&glinep,sfile,dest->page);

 return(NULL);
}




static os_error * linecopyc(objectloc * src,objectloc * dest,transformstr * tr)
{
 os_error  * err;
 int         gline;
 glinestr  * glinep;
 glinestr    srcline;
 glinestr    destline;

 err=NULL;

 glinep=(glinestr*)fob_find(src->object,src->file);
 srcline=*glinep;
 destline=srcline;

          err=copycurve(&destline.fcurve,dest->file,&srcline.fcurve,src->file);
 if(!err) err=fob_create(sizeof(glinestr),dest->file,OGLINE,(fobjptr)&glinep);
 if(!err)
 {
  gline=glinep->key;

  glinep->fcurve=destline.fcurve;

  if(src->branch && !dest->branch) glinep->master=srcline.key;

  if(dest->copymode & LOCOLDMASTER) glinep->master=srcline.master;
  else
  if(!src->branch && !dest->branch)
  {
   if(srcline.master && (dest->copymode & LOCKEEPMASTER))
   { 
    glinep->flags.local=destline.flags.local;
    mapmaster(src,dest,srcline.master,&destline.master);
    glinep=(glinestr*)fob_find(gline,dest->file);
    glinep->master=destline.master;
    glinep->flags.local=destline.flags.local;
    glinep->flags.masterdata=destline.flags.masterdata;
   }
  }

  tr_multiply(&glinep->transform,&destline.transform,tr);

  if(destline.fcurve.box.h) glinep->transform.f=0;
  else
  if(destline.fcurve.box.w) glinep->transform.e=0;

  getmaxbox(&glinep->box,&glinep->fcurve.box,&glinep->transform);
               /* dest->after,dest->parent */
  insafter(gline,0,dest->page,PGLINECHAIN,dest->file);

  dest->object=gline;

  vxadd((sobjectstr **)&glinep,dest->file);
 }
 return(err);
}


static os_error * linecopy(objectloc * src,objectloc * dest,transformstr * tr)
{
 if(dest->copymode & LOCMOVE) return(linecopym(src,dest,tr));
 else                         return(linecopyc(src,dest,tr));
}



static int findline(posnstr * posn,glinestr * glinep,int * before,int last)
{
 boxstr box;
 int    fx;
 int    fy;

 fx=fy=0;

 if(!glinep->fcurve.box.w) 
 {
  fx=2*ourunits(cvdu.deltax,posn->view);
 }
 else
 if(!glinep->fcurve.box.h)
 {
  fy=2*ourunits(cvdu.deltay,posn->view);
 }

 box=glinep->box;
 box.x-=fx;
 box.w+=2*fx;
 box.y-=fy;
 box.h+=2*fy;

 if(insidebox(posn->x,posn->y,&box))
 {
  last=glinep->key;
 }

 return(last);

 USE(before);
}



static int selectline(glinestr * glinep,posnstr * posn,int type)
{
 int           gline;
 int           parent;
 filestr     * file;

 gline=glinep->key;
 file=posn->file;

 if(type==SELECTOBJECT)
 {
  parent=parentof((chainobject*)glinep,file);
  if(parent) type=objectselect((objhdrstr*)fob_find(parent,file),posn,SELECT1);

  addselection(gline,file,type);
 }
 else
 if(type==DESELECTOBJECT)
 {
  remselection(gline,file,type);
 }

 return(type);
}



/* tagged to UGLTRAN */

static os_error * untransline(filestr * file)
{
 os_error      * err;
 int             gline;
 glinestr      * glinep;
 int             data[2];

 err=undoread(file,&gline,sizeof(int));

 glinep=(glinestr*)fob_find(gline,file);

 if(glinep)
 {
  vxadd((sobjectstr **)&glinep,file);

  data[0]=glinep->fcurve.box.x;
  data[1]=glinep->fcurve.box.y;

           err=undowritefn(file,UGLTRAN,&gline,sizeof(int));
  if(!err) err=undowrite(file,data,2*sizeof(int));
  if(!err) err=undoread(file,data,2*sizeof(int));

  fob_mod(gline,file,(fobjptr)&glinep);
  glinep->fcurve.box.x=data[0];
  glinep->fcurve.box.y=data[1];

  getmaxbox(&glinep->box,&glinep->fcurve.box,&glinep->transform);

  vxadd((sobjectstr **)&glinep,file);
 }

 return(err);
}





static os_error * transline(glinestr * glinep,filestr * file,
                                      transformfn trf,int flags)
{
 os_error     * err;
 transformstr   r1;
 int            data[2];
 int            gline;
 int            type;

 err=NULL;


 gline=glinep->key;

 if(!(flags & TRNOUPDATE)) vxadd((sobjectstr **)&glinep,file);

 if(!(flags & TRNOUNDO))
 {
  data[0]=glinep->fcurve.box.x;
  data[1]=glinep->fcurve.box.y;

           err=undowritefn(file,UGLTRAN,&gline,sizeof(int));
  if(!err) err=undowrite(file,data,2*sizeof(int));
 }

 fob_mod(gline,file,(fobjptr)&glinep);

 type=trf(&r1,&glinep->transform,&glinep->fcurve,&glinep->transform); 


 if((type & TR_SHIFT)==type)
 {
  if(!glinep->box.w) glinep->fcurve.box.x+=r1.e;
  else
  if(!glinep->box.h) glinep->fcurve.box.y+=r1.f;

  getmaxbox(&glinep->box,&glinep->fcurve.box,&glinep->transform);

  if(!(flags & TRNOUPDATE)) vxadd((sobjectstr **)&glinep,file);
 }

 return(err);
}


/*****************************************************************************/

static os_error *  linescrap(int object,filestr * file,int flags)
{
 os_error * err;
 glinestr * gline;

 err=NULL;

 gline=(glinestr*)fob_find(object,file);
 vxadd((sobjectstr **)&gline,file);

 if(flags & DELETEUNLINK) err=unlinkobject(object,PGLINECHAIN,file);
 if(!err) err=fob_delete(object,file);

 return(err);
}



static os_error *  uldeletesub(int object,filestr * file)
{
 os_error  * err;
 glinestr  * gline;
 int         parent;
 int         after;

 err=NULL;

 gline=(glinestr*)fob_find(object,file);
 if(gline)
 {
  vxadd((sobjectstr **)&gline,file);

  parent=gline->chain.parent;
  after=gline->chain.prev;

  unlinkobject(object,PGLINECHAIN,file);

  undowritefn(file,UGLRECREATE,&object,sizeof(int));
  undowrite(file,&parent,sizeof(int));
  undowrite(file,&after,sizeof(int));
 }

 return(err);
}


/* tagged to UGLRECREATE */

static os_error *  ulrecreate(filestr * file)
{
 glinestr * glinep;
 int        object;
 int        parent;
 int        after;

 undoread(file,&object,sizeof(int));
 undoread(file,&parent,sizeof(int));
 undoread(file,&after,sizeof(int));

 insafter(object,after,parent,PGLINECHAIN,file);

 undowritefn(file,UGLDELETE,&object,sizeof(int));

 glinep=(glinestr*)fob_find(object,file);

 vxadd((sobjectstr **)&glinep,file);

 return(NULL);
}



/* scrap function for UGLRECREATE */

static os_error * ultrash(filestr * file)
{
 int object;
 undoread(file,&object,sizeof(int));
 return(linescrap(object,file,0));
}



/* tagged to UGLDELETE */

static os_error *  uldelete(filestr * file)
{
 int object;
 undoread(file,&object,sizeof(int));
 uldeletesub(object,file);
 return(NULL);
}



static os_error * linedelete(glinestr * glinep,int dm,filestr * file)
{
 if(dm & DELETESCRAP) return(linescrap(glinep->key,file,dm));
 else                 return(uldeletesub(glinep->key,file));
}



/*****************************************************************************/


#define DSTGLINE        0x900


static void dsname(glinestr * glinep,char * data)
{
 sprintf(data,"GLINE_%02x",glinep->key);
}


static os_error * dsmessage(glinestr * glinep,filestr * file,buffer * bf)
{
 os_error * err;
 int        gline;

 gline=glinep->key;

 err=bf_printf(bf,"GLINE_%02x={pageguideline ",gline);

 if(glinep->master)
 {
  bf_printf(bf,"{master GLINE_%02x}{local %d}",glinep->master,
                                                 glinep->flags.local);
 }


 if(!glinep->master || glinep->flags.local)
 {
  bf_printfeol(bf,"{lock %d} {spread %d}",glinep->flags.locked,
                                         glinep->flags.spread);

  if(!glinep->box.w) bf_printf(bf,"{x %d}",glinep->box.x);
  else               bf_printf(bf,"{y %d}",glinep->box.y);
 }

 err=bf_printfeoln(bf,2,"}");
 
 return(err);

 USE(file);
}




static os_error * dsline(int fp,int argc,objectloc * loc)
{
 os_error    * err;
 int           sp;
 int           nsp;
 dsobjectstr * ds;
 dsobjectstr * ds1;
 int           master;
 int           local;
 int           coord;
 int           seenx;
 int           seeny;
 curvestr      c;
 glinestr    * glinep;
 objectloc     gloc;
 pagestr     * pagep;
 int           lock;
 int           spread;


 sp=fp+2; /* skip leading type/len */

 seenx=seeny=master=local=coord=lock=spread=0;

 while(argc--)
 {
  ds=(dsobjectstr*)(dsstack+sp);
  nsp=sp+ds->len/sizeof(int);
  ds1=(dsobjectstr*)(dsstack+sp+2);

  switch(ds->type)
  {
   case     DSTLOCAL:
                     local=ds1->data.integer;
                     break;

   case    DSTMASTER:
                     master=ds1->data.integer;
                     glinep=(glinestr*)fob_find(master,loc->file);
                     c=glinep->fcurve;
                     lock=glinep->flags.locked;
                     spread=glinep->flags.spread;
                     break;

   case      DSTLOCK:
                     lock=ds1->data.integer;
                     break;

   case    DSTSPREAD:
                     spread=ds1->data.integer;
                     break;

   case         DSTX:
                     coord=ds1->data.integer;
                     seenx=1;
                     break;

  case          DSTY:
                     coord=ds1->data.integer;
                     seeny=1;
                     break;
  }
  sp=nsp;
 }


 pagep=(pagestr*)fob_find(loc->page,loc->file);

 if(seenx || seeny)
 {
  c.type=CLINE;

  if(seenx)
  {
   c.box.h=pagep->box.h;
   c.box.w=0;
   c.box.x=coord;
   c.box.y=0;
  }
  else
  {
   c.box.h=0;
   c.box.w=pagep->box.w;
   c.box.x=0;
   c.box.y=coord;
  }
 }

 gloc=*loc;
 gloc.parent=loc->page;
 gloc.after=0;
 gloc.chain=PGLINECHAIN;

 err=createline(&gloc,&c);
 if(!err)
 {
  glinep=(glinestr*)fob_find(gloc.object,gloc.file);
  glinep->flags.local=local;
  glinep->master=master;
  glinep->flags.locked=lock;
  glinep->flags.spread=spread;

  ds=(dsobjectstr*)(dsstack+fp);
  ds->data.integer=gloc.object;
  ds->len=3*sizeof(int);
 }

 return(err);
}




static int dsblanklist[]={
                     DSTENDLIST
                    };




static dscreator dsdefs[]=
{
 0,0,0,0, "pageguideline",   dsline, 0, DSTGLINE,          dsblanklist,
};



/***************************************************************************/

static void pagesize(glinestr * glinep,filestr * file,sectiondatastr * data)
{
 if(!glinep->master || glinep->flags.local)
 {
  if(glinep->box.w) glinep->fcurve.box.w=data->width;
  else              glinep->fcurve.box.h=data->height;

  getmaxbox(&glinep->box,&glinep->fcurve.box,&glinep->transform);
 }

 USE(file);
}


static os_error * linemessage(glinestr * glinep,filestr * file,
                                                   int * message,void * data)

{
 omessagefn fn;
 snapfn     snapbox;

 switch(*message)
 {

  case   OM_FUNCTION:
  case  OM_FUNCTION2:
                     fn=((omessagefn)data);
                     return(fn((objhdrstr*)glinep,file,message));
                     break;

   case    OM_DSNAME:
                     dsname(glinep,(char*)data);
                     break;

   case    OM_DSSAVE:
                     return(dsmessage(glinep,file,(buffer*)data));
                     break;

  case OM_FINDMASTER:
                     if((*((int*)data))==glinep->master)
                     {
                      *((int*)data)=glinep->key;
                      *message=0;
                     }
                     break;


   case   OM_SNAPBOX:
                     if(!ignoresnapmessage)
                     {
                      snapbox=((snapfn)data);
                      snapbox(glinep->key,&glinep->transform,
                                          &glinep->fcurve.box);
                     }
                     break;


   case  OM_PAGESIZE:
                     pagesize(glinep,file,(sectiondatastr*)data);
                     break;

 }

 return(NULL);
}





static os_error * linedb(glinestr * glinep,filestr * file,
                                         viewstr * view,objectdbstr * cont)
{

 switch(cont->type)
 {
  case  ODB_OBJECTNAMEPLURAL:
                             tsprintf(ODBCONLEN,cont->name,"{LX03}");
                             break;

  case ODB_OBJECTNAMEPLURALC:
                             tsprintf(ODBCONLEN,cont->name,"{LX04}");
                             break;

  case            ODB_MODIFY:
                             openguidelinebox(glinep,file,view);
                             break;

  case        ODB_OBJECTNAME:
                             tsprintf(ODBCONLEN,cont->name,"{LX05}");
                             break;
 }

 return(NULL);
}



static os_error * masterobject(glinestr * glinep,filestr * file,int * master,
                                                          objectbitstr * bits)
{
 *master=glinep->master;
 memset(bits,0,sizeof(objectbitstr));
 bits->local=glinep->flags.local;
 bits->canmakelocal=1;
 bits->noresize=1;
 bits->norotate=1;
 bits->nogroup=1;
 bits->noshape=1;

 if(glinep->fcurve.box.h)
 {
  bits->vlock=1;
 }
 else
 {
  bits->hlock=1;
 }

 return(NULL);
 USE(file);
}


/***************************************************************************/
/* this is the trivial operation */

/* tagged to ULMXLOCAL */

static os_error *  unlmakelocal(filestr * file)
{
 os_error   * err;
 int          object;
 glinestr   * linep;

          err=undoread(file,&object,sizeof(int));
 if(!err) err=undowritefn(file,ULMXMASTER,&object,sizeof(int));
 if(!err) err=fob_mod(object,file,(fobjptr)&linep);
 if(!err)
 {
  linep->flags.local=1;
 }

 return(err);
}


/* tagged to ULMXMASTER */

static os_error *  unlmakemaster(filestr * file)
{
 os_error   * err;
 int          object;
 glinestr   * linep;

          err=undoread(file,&object,sizeof(int));
 if(!err) err=undowritefn(file,ULMXLOCAL,&object,sizeof(int));
 if(!err) err=fob_mod(object,file,(fobjptr)&linep);
 if(!err)
 {
  linep->flags.local=0;
 }

 return(err);
}




static os_error * makeobjectlocal(glinestr * linep,filestr * file)
{
 int master;

 master=linep->master;

 if(master)
 {
  if(!linep->flags.local)  
  {
   linep->flags.local=1;
   undowritefn(file,ULMXMASTER,&linep->key,sizeof(int));
  }
 }

 return(NULL);
}

/***************************************************************************/


static os_error * makemastersub(int linek,glinestr * xnew,filestr * file)
{
 os_error   * err;
 glinestr   * linep;

 err=fob_mod(linek,file,(fobjptr)&linep);

 vxadd((sobjectstr **)&linep,file);

 linep->transform=xnew->transform;
 linep->fcurve=xnew->fcurve;

 getmaxbox(&linep->box,&linep->fcurve.box,&linep->transform);

 vxadd((sobjectstr **)&linep,file);

 linep->flags.local=linep->flags.local ^ 1;

 return(err);
}


/* tagged to UMAKEMASTER */

static os_error * unmakemaster(filestr * file)
{
 os_error   * err;
 glinestr     xnew;
 glinestr     old;
 glinestr   * glinep;
 int          linek;

          err=undoread(file,&linek,sizeof(int));
 if(!err) err=undoread(file,&xnew.transform,sizeof(transformstr));
 if(!err) err=undoread(file,&xnew.fcurve,sizeof(curvestr));

 glinep=(glinestr*)fob_find(linek,file);
 old=*glinep;

 if(!err) err=undowritefn(file,UMAKEMASTER,&linek,sizeof(int));
 if(!err) err=undowrite(file,&old.transform,sizeof(transformstr));
 if(!err) err=undowrite(file,&old.fcurve,sizeof(curvestr));

 if(!err) err=makemastersub(linek,&old,file);

 return(err);
}



/* this is not trivial */

static os_error * makeobjectmaster(glinestr * linep,filestr * file)
{
 os_error  * err;
 int         master;
 glinestr    old;
 int         line;

 err=NULL;
 master=linep->master;

 if(master)
 {
  if(linep->flags.local)
  {
   line=linep->key;
   old=*linep;

            err=undowritefn(file,UMAKEMASTER,&line,sizeof(int));
   if(!err) err=undowrite(file,&old.transform,sizeof(transformstr));
   if(!err) err=undowrite(file,&old.fcurve,sizeof(curvestr));

   linep=(glinestr*)fob_find(master,file);
   old=*linep;

   makemastersub(line,&old,file);
  }
 }
 return(err);
}






static os_error * forceobjectmaster(int mode,glinestr * linep,filestr * file)
{
 os_error  * err;
 int         master;
 glinestr    old;
 glinestr    xnew;
 int         line;

 err=NULL;


 if(mode==FORCEMASTER || mode==FORCEMASTERU || mode==FORCEMASTERR)
 {
  line=linep->key;

  master=linep->master;

  if(mode==FORCEMASTERU) master=findoldmaster(master,file,1);

  if(master)
  {
   vxadd((sobjectstr **)&linep,file);

   old=*linep;

   deletecurve(&old.fcurve,file);

   linep=(glinestr*)fob_find(master,file);
   xnew=*linep;

   copycurve(&old.fcurve,file,&xnew.fcurve,file);

   err=fob_mod(line,file,(fobjptr)&linep);

   linep->transform=xnew.transform;
   linep->box=xnew.box;
   linep->fcurve=old.fcurve;

   vxadd((sobjectstr **)&linep,file);
  }
 }

 return(err);
}





static os_error * uudm(filestr * file)
{
 mrefstr      mref;
 objectloc    src;
 objectloc    dest;
 transformstr tr;

 undoread(file,&mref,sizeof(mrefstr));
 mpageundelete(&mref,PGLINECHAIN,&src,&dest,&tr,file);
 linecopyc(&src,&dest,&tr);
 return(undowritefn(file,URDM,&dest.object,sizeof(int)));
}


static os_error * deletemaster(glinestr * linep,filestr * file)
{
 int     line;
 mrefstr mref;

 line=linep->key;

 mref.master=linep->master;
 mref.after=linep->chain.prev;

 linescrap(line,file,DELETEUNLINK);

 return(undowritefn(file,UUDM,&mref,sizeof(mrefstr)));
}


static os_error * urdm(filestr * file)
{
 int        line;
 glinestr * linep;

 undoread(file,&line,sizeof(int));
 linep=(glinestr*)fob_find(line,file);
 return(deletemaster(linep,file));
}





static os_error * localobject(glinestr * glinep,filestr * file,int makelocal)
{
 if(makelocal==MAKELOCAL)    return(makeobjectlocal(glinep,file));
 else
 if(makelocal==MAKEMASTER)   return(makeobjectmaster(glinep,file));
 else
 if(makelocal==DELETEMASTER) return(deletemaster(glinep,file));
 else                        return(forceobjectmaster(makelocal,glinep,file));
}





static os_error * oreaddata(int action,glinestr * glinep,
                                             filestr * file,void * data)

{

 switch(action)
 {
  case SELECTINFO:
                  ((outlineinfostr*)data)->fcurve=glinep->fcurve;
                  ((outlineinfostr*)data)->tr=glinep->transform;
                  ((outlineinfostr*)data)->frame=0;
                  break;

 }

 return(NULL);

 USE(file);
}




/****************************************************************************/


static objectstr lineobject=
{
 OGLINE,                          /* type */
 NULL,                            /* new  */
 (objectcopyfn)linecopy,          /* copy */
 (objectredrawfn)redrawline,      /* redraw */
 (objectcreatefn)createline,      /* create */
 (objectfindfn)findline,          /* find   */
 (objectselectfn)selectline,      /* select */
 (objecttransformfn)transline,    /* transform */
 NULL,                            /* load type */
 NULL,                            /* load */
 NULL,                            /* load post */
 (objectmessagefn)linemessage,    /* message */
 (objectdbfn)linedb,              /* object db */
 NULL,                            /* text flow       */
 (objectdeletefn)linedelete,      /* delete          */
 NULL,                            /* text reflow     */
 NULL,                            /* object set flow */
 (objectmasterfn)masterobject,    /* master object?  */
 (objectlocalfn)localobject,      /* local object    */
 NULL,                            /* change shape    */
 (objectreadfn)oreaddata,         /* read data */
 NULL,                            /* x12 */

};


/***************************************************************************/

static pagesectionstr rpagesection;

/* tagged to URDELETES */
/* this function is always called last in undo/redo */

static os_error * undeleterulers(filestr * file) 
{
 vxend(file,VRBACK,rpagesection.page,rpagesection.section);

 return(undowritefn(file,URDELETEF,&rpagesection,sizeof(pagesectionstr)));
}


/* tagged to URDELETEF */

static os_error * undeleterulerf(filestr * file)
{
 vxstart(file);

 undoread(file,&rpagesection,sizeof(pagesectionstr));

 return(undowritefn(file,URDELETES,NULL,0));
}


os_error * rulerdelete(filestr * file,int h,int page,int sect)
{
 os_error   * err;
 int          master;
 objectbitstr bits;
 pagestr   *  pagep;
 int          gline;
 glinestr  *  glinep;
 int          next;

 err=NULL;

 pagep=(pagestr*)fob_find(page,file);
 gline=pagep->gchain.child;

 if(gline)
 {
  err=undostart(file);
  if(!err) err=modfile(file,sect);
  if(!err)
  {
   if(!err)
   {
    vxstart(file);
    undowritefn(file,URDELETES,NULL,0);
    rpagesection.page=page;
    rpagesection.section=sect;

    while(gline)
    {
     glinep=(glinestr*)fob_find(gline,file);
     next=glinep->chain.next;

     if(( h && !glinep->fcurve.box.w) ||
        (!h && !glinep->fcurve.box.h))
     {
      objectmaster(gline,file,&master,&bits);
      if(!master || bits.local)
      {
       if(isselected(gline,file)) remselection(gline,file,DESELECTOBJECT);

       err=objectdelete((objhdrstr*)fob_find(gline,file),0,file);
      }
     }
     gline=next;
    }

    undeleterulers(file);
   }
  }
  undoend(file);
 }
 return(err);
}




/****************************************************************************/

/* tagged to URCREATES */
/* this function is always called last in undo/redo */

static os_error * uncreates(filestr * file) 
{
 vxend(file,VRBACK,rpagesection.page,rpagesection.section);
 return(undowritefn(file,URCREATEF,&rpagesection,sizeof(pagesectionstr)));
}


/* tagged to URCREATEF */

static os_error * uncreatef(filestr * file)
{
 undoread(file,&rpagesection,sizeof(pagesectionstr));
 vxstart(file);
 return(undowritefn(file,URCREATES,NULL,0));
}


os_error * rulercreate(posnstr * posn,int h)
{
 os_error * err;
 curvestr   c;
 objectloc  loc;
 pagestr  * pagep;


 pagep=(pagestr*)fob_find(posn->page,posn->file);

 c.type=CLINE;

 if(h)
 {
  c.box.h=pagep->box.h;
  c.box.w=0;
  c.box.x=posn->x;
  c.box.y=0;
 }
 else
 {
  c.box.h=0;
  c.box.w=pagep->box.w;
  c.box.x=0;
  c.box.y=posn->y;
 }

 loc.file=posn->file;
 loc.object=0;
 loc.section=posn->section;
 loc.page=posn->page;
 loc.parent=posn->page;
 loc.after=0;
 loc.branch=0;
 loc.chain=PGLINECHAIN;

 err=modfile(loc.file,loc.section);
 if(!err)
 {
  undostart(loc.file);

  undowritefn(loc.file,URCREATES,NULL,0);
  vxstart(loc.file);

  err=objectcreate(&loc,OGLINE,&c);
  rpagesection.page=loc.page;
  rpagesection.section=loc.section;
  posn->object=loc.object;

  uncreates(loc.file);

  undoend(loc.file);
 }

 return(err);
}

/****************************************************************************/


static int           seenxl;
static int           seenxr;
static int           seenyt;
static int           seenyb;
static int           xnl;
static int           xnr;
static int           ynt;
static int           ynb;
static int           xl;
static int           xr;
static int           yt;
static int           yb;

static int         * xplp;
static int         * xprp;
static int         * yptp;
static int         * ypbp;

static int           snapobject;



static void snapfunction(int object,transformstr * tr,boxstr * sbox)
{
 boxstr box;
 int    xtemp;
 int    ytemp;


 if(object!=snapobject)
 {
  if(tr) getmaxbox(&box,sbox,tr);
  else   box=*sbox;

  if(xplp)
  {
   xtemp=box.x;

   if(seenxl)
   {
    if((ABS(xtemp-xl))<(ABS(xl-xnl))) xnl=xtemp;
   }
   else
   {
    seenxl=1;
    xnl=xtemp;
   }
  }

  if(xprp)
  {
   xtemp=box.x+box.w;

   if(seenxr)
   {
    if((ABS(xtemp-xr))<(ABS(xr-xnr))) xnr=xtemp;
   }
   else
   {
    seenxr=1;
    xnr=xtemp;
   }
  }

  if(yptp)
  {
   ytemp=box.y;
 
   if(seenyt)
   {
    if((ABS(ytemp-yt))<(ABS(yt-ynt))) ynt=ytemp;
   }
   else
   {
    seenyt=1;
    ynt=ytemp;
   }
  }
 
  if(ypbp)
  {
   ytemp=box.y+box.h;

   if(seenyb)
   {
    if((ABS(ytemp-yb))<(ABS(yb-ynb))) ynb=ytemp;
   }
   else
   {
    seenyb=1;
    ynb=ytemp;
   }
  }
 }
}



static void snaptoframes(int object,posnstr * posn,filestr * sfile,
                                                   filestr * dfile)
{
 int            message;

 if(sfile!=dfile) snapobject=0;
 else             snapobject=posn->object;

 /* problem here is that snaps to glines are done elsewhere */
 /* and because we can switch them off, we don't want them  */

 ignoresnapmessage=1; 
 message=OM_SNAPBOX;
 objectmessage2(object,dfile,&message,(void*)snapfunction);
 ignoresnapmessage=0;

 return;

#ifdef NEVER

 while(object)
 {
  sob=fob_find(object,dfile);
  if(sob->type==OPAGE) object=0;
  else                 object=sob->chain.next;

  if(sob->key!=posn->object || sfile!=dfile)
  {
   message=OM_SNAPBOX;
   objectmessage((objhdrstr*)sob,dfile,&message,(void*)snapfunction);
   snaptoframes(sob->chain.child,posn,sfile,dfile);
  }
 }

#endif

}



void snap(int * xpl,int * xpr,int * ypt,int * ypb,
                                        posnstr * posn,posnstr * curr)
{
 prefdatastr * pref;
 filestr     * sfile;
 filestr     * dfile;
 viewstr     * view;
 int           dpage;
 int           snaprulerguidelines;
 int           snapframes;
 pagestr     * pagep;
 glinestr    * glinep;
 int           gline;
 sobjectstr  * sob;
 int           sxr;
 int           syr;


 seenxl=seenxr=seenyt=seenyb=0;

 xl=xr=yt=yb=0;

 xnl=xnr=ynt=ynb=0;

 xplp=xpl;
 xprp=xpr;
 yptp=ypt;
 ypbp=ypb;


 if(xpl) xl=*xpl;
 if(xpr) xr=*xpr;
 if(ypt) yt=*ypt;
 if(ypb) yb=*ypb;


 sfile=dfile=posn->file;
 view=posn->view;
 dpage=posn->page;
 if(curr) 
 {
  dfile=curr->file;
  view=curr->view;
  dpage=curr->page;
 }


 pref=prefp(dfile);
 snaprulerguidelines=pref->snaprulerguidelines;
 snapframes=pref->snapframes;

 if(posn->object)
 {
  sob=(sobjectstr*)fob_find(posn->object,sfile);
  if(sob->type==OGLINE) snaprulerguidelines=snapframes=0;
 }

 if(view->nosnap) snaprulerguidelines=snapframes=0;

 if(snaprulerguidelines)
 {
  pagep=(pagestr*)fob_find(dpage,dfile);

  gline=pagep->gchain.child;

  while(gline)
  {
   glinep=(glinestr*)fob_find(gline,dfile);
   gline=glinep->chain.next;

   if(glinep->fcurve.box.h)
   {
    if(seenxl)
    {
     if(xpl && (ABS(glinep->box.x-xl))<(ABS(xl-xnl))) xnl=glinep->box.x;
     if(xpr && (ABS(glinep->box.x-xr))<(ABS(xr-xnr))) xnr=glinep->box.x;
    }
    else
    {
     seenxl=seenxr=1;
     xnl=xnr=glinep->box.x;
    }
   }
   else
   {
    if(seenyt)
    {
     if((ABS(glinep->box.y-yt))<(ABS(yt-ynt))) ynt=glinep->box.y;
     if((ABS(glinep->box.y-yb))<(ABS(yb-ynb))) ynb=glinep->box.y;
    }
    else
    {
     seenyt=seenyb=1;
     ynt=ynb=glinep->box.y;
    }
   }
  }
 }

 if(snapframes) snaptoframes(dpage,posn,sfile,dfile);

 sxr=ourunits(snapxrange,view);
 syr=ourunits(snapyrange,view);

 if(seenxl && (ABS(xl-xnl))>sxr) seenxl=0;
 if(seenxr && (ABS(xr-xnr))>sxr) seenxr=0;
 if(seenyt && (ABS(yt-ynt))>syr) seenyt=0;
 if(seenyb && (ABS(yb-ynb))>syr) seenyb=0;

 if(xpl && xpr && (seenxl || seenxr))
 {
  if((ABS(xl-xnl))>(ABS(xr-xnr)))
  {
   *xpr=xnr;
   *xpl=xnr-(xr-xl);
  }
  else
  {
   *xpl=xnl;
   *xpr=xnl+(xr-xl);
  }
 }
 else
 if(xpl && seenxl)
 {
  *xpl=xnl;
 }
 else
 if(xpr && seenxr)
 {
  *xpr=xnr;
 }

 if(ypt && ypb && (seenyt || seenyb))
 {
  if((ABS(yt-ynt))>(ABS(yb-ynb)))
  {
   *ypb=ynb;
   *ypt=ynb-(yb-yt);
  }
  else
  {
   *ypt=ynt;
   *ypb=ynt+(yb-yt);
  }
 }
 else
 if(ypt && seenyt)
 {
  *ypt=ynt;
 }
 else
 if(ypb && seenyb)
 {
  *ypb=ynb;
 }

}



void snap1(int * xp,int * yp,posnstr * posn)
{
 int x;
 int xl;
 int xr;
 int y;
 int yt;
 int yb;

 x=xl=xr=*xp;
 y=yt=yb=*yp;

 snap(&xl,&xr,&yt,&yb,posn,NULL);

 if((ABS(x-xl))<(ABS(x-xr))) *xp=xl;
 else                        *xp=xr;

 if((ABS(y-yt))<(ABS(y-yb))) *yp=yt;
 else                        *yp=yb;
}


/****************************************************************************/

#define UNDOS 13

static undofnstr undos[UNDOS]=
{
 untransline,   NULL,    UGLTRAN,     "{UTRANS}",
 ulrecreate,    ultrash, UGLRECREATE, NULL,
 uldelete,      NULL,    UGLDELETE,   NULL,

 undeleterulers,NULL,    URDELETES,   "{UDGLINE}",
 undeleterulerf,NULL,    URDELETEF,   "{UDGLINE}",

 uncreates,     NULL,    URCREATES,   "{UCGLINE}",
 uncreatef,     NULL,    URCREATEF,   "{UCGLINE}",

 unlmakelocal,  NULL,    ULMXLOCAL,   NULL,
 unlmakemaster, NULL,    ULMXMASTER,  NULL,
 unmakemaster,  NULL,    UMAKEMASTER, NULL,

 unlinemove,    NULL,    UGMOVE,      NULL,

 uudm,          NULL,    UUDM,        NULL,
 urdm,          NULL,    URDM,        NULL,

};


os_error * guidelineinit(void)
{
 os_error * err;
 err=undoaddblock(undos,UNDOS);
 if(!err) err=objectregister(&lineobject);
 dsaddblock(dsdefs,sizeof(dsdefs)/sizeof(dscreator));
 return(err);
}


