/*->c.group */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"
#include "xprocess.h"


#include "err.h"
#include "xmath.h"
#include "fsx.h"
#include "bf.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "colour.h"
#include "etc.h"


#include "constants.h"
#include "str.h"
#include "object.h"
#include "reslink.h"

#include "viewr.h"
#include "view.h"
#include "file.h"
#include "chain.h"
#include "xredraw.h"
#include "geometry.h"
#include "curve.h"
#include "posn.h"
#include "xcolour.h"

#include "select.h"
#include "objectdb.h"
#include "undo.h"
#include "dscript.h"

#include "mpage.h"


#include "frameflow.h"


#include "group.h"
#include "dbgroup.h"



#define UGTRAN     0x5000
#define UBACK      0x5001
#define UFORWARD   0x5002
#define UGDELETE   0x5003
#define UGRECREATE 0x5004
#define UGDELINK   0x5005
#define UGRESTORE  0x5006
#define ULMXMASTER 0x5007
#define ULMXLOCAL  0x5008
#define UGMASTERS  0x5009
#define UGMASTERF  0x500A
#define UGMOVE     0x500B
#define UUDM       0x500C
#define URDM       0x500D
#define UGUPDATE   0x500E




static os_error * groupredraw(viewredrawstr * v,groupstr * groupp)
{
 os_error * err;
 int        object;

 err=NULL;

 object=groupp->chain.child;
 while(object)
 {
  groupp=(groupstr*)fob_find(object,v->file);
  object=groupp->chain.next;
  err=objectredraw(v,(objhdrstr*)groupp);
  if(err) break;
 }

 return(err);
}






static os_error * groupcreate(objectloc * loc,curvestr * c)
{
 os_error   * err;
 int          groupk;
 groupstr   * groupp;
 curvestr     newi;
 curvestr     newo;
 curvestr     fcurve;

 err=fob_create(sizeof(groupstr),loc->file,OGROUP,(fobjptr)&groupp);
 if(!err)
 {
  groupk=groupp->key;

  tr_setidentity(&groupp->transform);

  copycurve(&fcurve,loc->file,c,loc->file);
  newcurve(&newi,CBOX,NULL,0,loc->file);          /* not using inner curve */
  newcurve(&newo,CBOX,NULL,0,loc->file);          /* not using outer curve */

  groupp=(groupstr*)fob_find(groupk,loc->file);

  groupp->fcurve=fcurve;
  groupp->icurve=newi;
  groupp->ocurve=newo;

  getmaxbox(&groupp->box,&groupp->fcurve.box,&groupp->transform);

  insafter(groupk,loc->after,loc->parent,loc->chain,loc->file);

  loc->object=groupk;

  undowritefn(loc->file,UGDELETE,&loc->object,sizeof(int));
 }
 return(err);
}





/*****************************************************************************/
/* writes data that can be changed by a transform */

static os_error * writetransdata(groupstr * groupp,filestr * file)
{
 os_error     * err;
 err=undowrite(file,&groupp->transform,sizeof(transformstr));
 return(err);
}

/* reads data that can be changed by a transform */

static os_error * readtransdata(groupstr * groupp,filestr * file)
{
 os_error     * err;
 err=undoread(file,&groupp->transform,sizeof(transformstr));
 return(err);
}


/* tagged to UGTRAN */

static os_error * ungrouptransform(filestr * file)
{
 os_error     * err;
 int            group;
 groupstr     * groupp;

 err=undoread(file,&group,sizeof(int));

 groupp=(groupstr*)fob_find(group,file);

 if(groupp)
 {
           err=undowritefn(file,UGTRAN,&group,sizeof(int));
  if(!err) err=writetransdata(groupp,file);
  if(!err) err=readtransdata(groupp,file);

  getmaxbox(&groupp->box,&groupp->fcurve.box,&groupp->transform);
 }

 return(err);
}



static transformstr gtrans;


static int gtrf(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct)
{
 tr_multiply(r,t,&gtrans);
 return(0);
 USE(c);
 USE(ct);
}


static os_error * grouptransform(groupstr * groupp,filestr * file,
                                   transformfn trf,int flags)
{
 os_error     * err;
 transformstr   inv;
 transformstr   r1;
 int            code;
 int            group;
 int            object;
 sobjectstr   * sob;

 group=groupp->key;

 err=NULL;

 if(!(flags & TRNOUNDO))
 {
           err=undowritefn(file,UGTRAN,&group,sizeof(int));
  groupp=(groupstr*)fob_find(group,file);
  if(!err) err=writetransdata(groupp,file);
 }

 groupp=(groupstr*)fob_find(group,file);
 code=trf(&r1,&groupp->transform,&groupp->fcurve,&groupp->transform); 
 tr_inverse(&groupp->transform,&inv);
 tr_multiply(&gtrans,&inv,&r1);

 groupp->transform=r1;

 getmaxbox(&groupp->box,&groupp->fcurve.box,&groupp->transform);

 groupp=(groupstr*)fob_find(group,file);
 object=groupp->chain.child;
 while(object)
 {
  sob=(sobjectstr*)fob_find(object,file);
  object=sob->chain.next;

  err=objecttransform((objhdrstr*)sob,file,gtrf,flags);
  if(err) break;
 }

 return(err);
}


/****************************************************************************/

os_error * groupupdate(int group,int lock,int spread,filestr * file)
{
 os_error      * err;
 groupstr      * groupp;
 flagstr         flags;

 err=NULL;

 groupp=(groupstr*)fob_find(group,file);
 flags=groupp->flags;

 undowritefn(file,UGUPDATE,&group,sizeof(int));
 undowrite(file,&flags,sizeof(flagstr));

 groupp=(groupstr*)fob_find(group,file);
 groupp->flags.locked=lock;
 groupp->flags.spread=spread;

 return(err);
}

static os_error * ungroupupdate(filestr * file)
{
 flagstr         flags;
 int             group;

 undoread(file,&group,sizeof(int));
 undoread(file,&flags,sizeof(flagstr));
 return(groupupdate(group,flags.locked,flags.spread,file));
}


/****************************************************************************/

/* move an object - within the same file */

static transformstr xtransform;

static int xtrf(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct)
{
 tr_multiply(r,t,&xtransform);

 return(0);

 USE(c);
 USE(ct);
}



typedef struct grxmovestr
{
 int            group;
 int            spage;
 int            dpage;
 int            parent;
 int            after;
 transformstr   tr;
} grxmovestr;




static os_error * ungmove(filestr * file)
{
 grxmovestr    old;
 grxmovestr    xnew;
 groupstr    * groupp;

 undoread(file,&xnew,sizeof(grxmovestr));

 old=xnew;
 old.spage=xnew.dpage;
 old.dpage=xnew.spage;

 groupp=(groupstr*)fob_find(old.group,file);

 old.tr=groupp->transform;
 old.after=groupp->chain.prev;
 old.parent=parentof((chainobject*)groupp,file);

 tr_inverse(&xnew.tr,&old.tr);

 xtransform=old.tr;

 undowritefn(file,UGMOVE,&old,sizeof(grxmovestr));

 groupp=(groupstr*)fob_find(xnew.group,file);

 vxaddn((sobjectstr**)&groupp,file,xnew.spage);
 frameflow((sobjectstr **)&groupp,file);

 unlinkobject(old.group,0,file);
 insafter(xnew.group,xnew.after,xnew.parent,0,file);

 groupp=(groupstr*)fob_find(xnew.group,file);

 grouptransform(groupp,file,xtrf,TRNOUPDATE|TRNOREFLOWS|TRNOUNDO);

 vxaddn((sobjectstr **)&groupp,file,xnew.dpage);

 return(NULL);
}




static os_error * groupcopym(objectloc * src,
                             objectloc * dest,transformstr * tr)
{
 groupstr   * groupp;
 filestr    * sfile;
 int          object;
 int          trmode;
 grxmovestr   old;


 sfile=src->file;
 object=src->object;

 groupp=(groupstr*)fob_find(object,sfile);
 vxaddn((sobjectstr**)&groupp,sfile,src->page);
 frameflow((sobjectstr **)&groupp,sfile);

 if(dest->copymode & LOCUNDO)
 {
  old.group=object;
  old.spage=dest->page;
  old.dpage=src->page;
  old.tr=*tr;
  old.after=groupp->chain.prev;
  old.parent=parentof((chainobject*)groupp,sfile);

  undowritefn(sfile,UGMOVE,&old,sizeof(grxmovestr));
 }

 unlinkobject(object,0,sfile);
 dest->object=object;
 insafter(object,dest->after,dest->parent,0,sfile);

 groupp=(groupstr*)fob_find(object,sfile);
 xtransform=*tr;

 trmode=TRNOUPDATE|TRNOREFLOWS|TRNOUNDO;

 grouptransform(groupp,sfile,xtrf,trmode);

 groupp=(groupstr*)fob_find(object,sfile);
 vxaddn((sobjectstr**)&groupp,sfile,dest->page);

 return(NULL);
}




static os_error * groupcopyc(objectloc * src,
                             objectloc * dest,transformstr * tr)
{
 os_error    * err;
 int           groupk;
 groupstr    * groupp;
 groupstr      srcgp;
 groupstr      destgp;
 objectloc     gsrc;
 objectloc     gdest;
 sobjectstr  * sob;
 filestr     * sfile;
 filestr     * dfile;


 err=NULL;

 sfile=src->file;
 dfile=dest->file;

 groupp=(groupstr*)fob_find(src->object,sfile);
 srcgp=*groupp;
 destgp=srcgp;

          err=copycurve(&destgp.fcurve,dfile,&srcgp.fcurve,sfile);
 if(!err) err=copycurve(&destgp.icurve,dfile,&srcgp.icurve,sfile);
 if(!err) err=copycurve(&destgp.ocurve,dfile,&srcgp.ocurve,sfile);


 if(!err) err=fob_create(sizeof(groupstr),dfile,OGROUP,(fobjptr)&groupp);
 if(!err)
 {
  groupk=groupp->key;

  groupp->fcurve=destgp.fcurve;
  groupp->icurve=destgp.icurve;
  groupp->ocurve=destgp.ocurve;
  groupp->textflow=destgp.textflow;

  if(src->branch && !dest->branch) groupp->master=srcgp.key;

  if(dest->copymode & LOCOLDMASTER) groupp->master=srcgp.master;
  else
  if(!src->branch && !dest->branch)
  {
   if(srcgp.master && (dest->copymode & LOCKEEPMASTER))
   { 
    groupp->flags.local=destgp.flags.local;
    mapmaster(src,dest,srcgp.master,&destgp.master);
    groupp=(groupstr*)fob_find(groupk,dfile);
    groupp->master=destgp.master;
    groupp->flags.local=destgp.flags.local;
   }
  }

  tr_multiply(&groupp->transform,&destgp.transform,tr);

  getmaxbox(&groupp->box,&groupp->fcurve.box,&groupp->transform);

  insafter(groupk,dest->after,dest->parent,dest->chain,dfile);

  dest->object=groupk;

  /* now have to recursively call objectcopy with our own source and dest */

  gsrc.object=srcgp.chain.child;
  gsrc.parent=srcgp.key;
  gsrc.after=0;

  gdest.parent=groupk;
  gdest.after=0;

  while(gsrc.object)
  {
   gsrc.file=sfile;
   gsrc.section=src->section;
   gsrc.page=src->page;
   gsrc.chain=0;
   gsrc.branch=src->branch;
   gsrc.copymode=0;

   gdest.file=dfile;
   gdest.object=0;
   gdest.section=dest->section;
   gdest.page=dest->page;
   gdest.chain=0;
   gdest.branch=dest->branch;
   gdest.copymode=dest->copymode;

   err=objectcopy(&gsrc,&gdest,tr);
   if(err) break;

   gdest.after=gdest.object;
   gdest.parent=0;

   sob=(sobjectstr*)fob_find(gsrc.object,sfile);
   gsrc.after=gsrc.object;
   gsrc.parent=0;
   gsrc.object=sob->chain.next;
  }
 }

 return(err);
}



static os_error * groupcopy(objectloc * src,objectloc * dest,transformstr * tr)
{
 if(dest->copymode & LOCMOVE) return(groupcopym(src,dest,tr));
 else                         return(groupcopyc(src,dest,tr));
}


/*****************************************************************************/



static int groupfind(posnstr * posn,groupstr * groupp,int * before,int last)
{
 int           groupk;
 int           childk;
 chainobject * childp;

 groupk=groupp->key;

 if(insidebox(posn->x,posn->y,&groupp->box))
 {
  if(insidecurve(posn->x,posn->y,&groupp->fcurve,&groupp->transform,
                                                  posn->file)) last=groupk;
 }

 childk=groupp->chain.child;
 while(childk)
 {
  childp=(chainobject*)fob_find(childk,posn->file);
  childk=childp->chain.next;
  last=objectfind(posn,(objhdrstr*)childp,before,last);
 }
 return(last);
}



static int groupselect(groupstr * groupp,posnstr * posn,int type)
{
 int           groupk;
 int           parent;
 filestr     * file;

 groupk=groupp->key;
 file=posn->file;

 if(type==SELECTOBJECT || type==SELECT1)
 {
  parent=parentof((chainobject*)groupp,file);
  if(parent) type=objectselect((objhdrstr*)fob_find(parent,file),posn,SELECT1);
  addselection(groupk,file,type);
  if(type==SELECT1) type=SELECT0;
 }
 else
 if(type==DESELECTOBJECT)
 {
  remselection(groupk,file,type);
 }

 return(type);
}

/*****************************************************************************/


static os_error * groupscrap(int object,filestr * file,int flags)
{
 os_error   * err;
 groupstr     group;
 int          child;
 sobjectstr * sob;

 if(flags & DELETEUNLINK) unlinkobject(object,0,file);

 group=*((groupstr*)fob_find(object,file));

 child=group.chain.child;

 while(child)
 {
  sob=(sobjectstr*)fob_find(child,file);
  child=sob->chain.next;
  err=objectdelete((objhdrstr*)sob,flags,file);
  if(err) break;
 }

          err=deletecurve(&group.fcurve,file);
 if(!err) err=deletecurve(&group.icurve,file);
 if(!err) err=deletecurve(&group.ocurve,file);

 fob_delete(group.key,file);

 return(err);
}



static os_error * ugdeletesub(int object,filestr * file)
{
 os_error   * err;
 groupstr   * group;
 int          parent;
 int          after;
 int          child;
 sobjectstr * sob;


 err=NULL;

 group=(groupstr*)fob_find(object,file);
 if(group)
 {
  parent=group->chain.parent;
  after=group->chain.prev;
  child=group->chain.child;

  while(child)
  {
   sob=(sobjectstr*)fob_find(child,file);
   child=sob->chain.next;
   err=objectdelete((objhdrstr*)sob,0,file);
   if(err) break;
  }

  unlinkobject(object,0,file);

  undowritefn(file,UGRECREATE,&object,sizeof(int));
  undowrite(file,&parent,sizeof(int));
  undowrite(file,&after,sizeof(int));
 }

 return(err);
}




/* tagged to UGRECREATE */

static os_error *  ugrecreate(filestr * file)
{
 int object;
 int parent;
 int after;

 undoread(file,&object,sizeof(int));
 undoread(file,&parent,sizeof(int));
 undoread(file,&after,sizeof(int));

 insafter(object,after,parent,0,file);

 undowritefn(file,UGDELETE,&object,sizeof(int));

 return(NULL);
}



/* scrap function for UGRECREATE */

static os_error * ugtrash(filestr * file)
{
 int object;
 undoread(file,&object,sizeof(int));
 return(groupscrap(object,file,DELETESCRAP));
}




/* tagged to UGDELETE */

static os_error * ugdelete(filestr * file)
{
 int object;
 undoread(file,&object,sizeof(int));
 ugdeletesub(object,file);
 return(NULL);
}



static os_error * groupdelete(groupstr * groupp,int dm,filestr * file)
{
 if(dm & DELETESCRAP) return(groupscrap(groupp->key,file,dm));
 else                 return(ugdeletesub(groupp->key,file));
}


/****************************************************************************/

static os_error * groupsize(int group,filestr * file)
{
 os_error   * err;
 groupstr   * groupp;
 int          child;
 sobjectstr * sob;
 transformstr inv;
 transformstr ivx;
 aboxstr      maxbox;
 boxstr       temp1;
 aboxstr      temp2;
 int          first;

 err=NULL;

 groupp=(groupstr*)fob_find(group,file);
 if(groupp)
 {
  first=1;

  tr_inverse(&groupp->transform,&inv);

  child=groupp->chain.child;
  while(child)
  {
   sob=(sobjectstr*)fob_find(child,file);
   child=sob->chain.next;

   tr_multiply(&ivx,&sob->transform,&inv);

   getmaxbox(&temp1,&sob->fcurve.box,&ivx);
   box2abox(&temp1,&temp2);

   if(first) 
   {
    maxbox=temp2;
    first=0;
   }
   else
   {
    getsumabox(&maxbox,&temp2);
   }
  }

  abox2box(&maxbox,&temp1);


  fob_mod(group,file,(fobjptr)&groupp);

  groupp->fcurve.box=temp1;
  getmaxbox(&groupp->box,&groupp->fcurve.box,&groupp->transform);
 }
 return(err);
}






/****************************************************************************/

#define DSTGROUP 0x1100


static os_error * dsgroup(int fp,int argc,objectloc * loc)
{
 os_error    * err;
 filestr     * file;
 dsobjectstr * ds;
 dsobjectstr * ds1;
 int           nsp;
 int           sp;
 groupstr    * groupp;
 int           group;
 int           object;
 int           parent;
 int           after;
 int           local;
 int           master;


 file=loc->file;

 err=fob_create(sizeof(groupstr),loc->file,OGROUP,(fobjptr)&groupp);
 if(!err)
 {
  group=groupp->key;
  tr_setidentity(&groupp->transform);
  newcurve(&groupp->fcurve,CBOX,NULL,0,loc->file); 

  insafter(group,loc->after,loc->parent,loc->chain,file);

  sp=fp+2; /* skip leading type/len */

  after=0;
  parent=group;
  local=master=0;

  while(argc--)
  {
   ds=(dsobjectstr*)(dsstack+sp);
   nsp=sp+ds->len/sizeof(int);
   ds1=(dsobjectstr*)(dsstack+sp+2);


   switch(ds->type)
   {
    case     DSTLOCAL:
                      local=ds1->data.integer;
                      break;

    case    DSTMASTER:
                      master=ds1->data.integer;
                    /*  groupp=fob_find(master,loc->file);
                      lock=groupp->flags.locked;
                      spread=grpupp->flags.spread; */
                      break;

              default:
                      object=ds->data.integer;

                      unlinkobject(object,0,file);
                      insafter(object,after,parent,0,file);
                      after=object;
                      parent=0;
                      break;
   }

   sp=nsp;
  }

  groupsize(group,file);

  groupp=(groupstr*)fob_find(group,loc->file);
  groupp->flags.local=local;
  groupp->master=master;

  loc->after=group;
  loc->parent=0;
  loc->object=group;

  ds=(dsobjectstr*)(dsstack+fp);
  ds->data.integer=group;
  ds->len=3*sizeof(int);
 }

 return(NULL);
}





static int dsblanklist[]={
                     DSTENDLIST
                    };




static dscreator dsdefs[]=
{
 0,0,0,0, "group",  dsgroup,   0, DSTGROUP,     dsblanklist,
};



static os_error * groupdssave(groupstr * groupp,int * message,
                                                filestr * file,buffer * bf)
{
 int group;

 group=groupp->key;

 objectchainmessage(groupp->chain.child,file,message,bf);

 bf_printf(bf,"GROUP_%02x={group ",group);

 if(groupp->master)
 {
  bf_printf(bf,"{master GROUP_%02x}{local %d} ",groupp->master,
                                                groupp->flags.local);
 }


 groupp=(groupstr*)fob_find(group,file);

 dschain(groupp->chain.child,file,bf);

 return(bf_printfeoln(bf,2,"}"));
}



static os_error * groupmessage(groupstr * groupp,filestr * file,
                                                   int * message,void * data)

{
 os_error * err;
 int        group;
 int        parent;
 static int active;
 int        child;
 omessagefn fn;
 snapfn     snapbox;

 err=NULL;

 child=groupp->chain.child;

 switch(*message)
 {

  case    OM_GETFILL:
  case    OM_SETFILL:
                     return(NULL);


  case   OM_EDITPATH:             /* stop editing of objects in groups */
                     return(NULL);


  case   OM_FUNCTION:
  case  OM_FUNCTION2:
                     fn=((omessagefn)data);
                     err=fn((objhdrstr*)groupp,file,message);
                     break;


  case     OM_DSSAVE:
                     return(groupdssave(groupp,message,file,(buffer*)data));
                     break;


  case     OM_DSNAME:
                     sprintf((char*)data,"GROUP_%02x",groupp->key);
                     return(NULL);
                     break;


  case OM_CHANGESIZE:
                     if(!active)
                     {
                      active=1;
                      group=groupp->key;
                      parent=parentof((chainobject*)groupp,file);
                      groupsize(group,file);
                      objectmessage((objhdrstr*)fob_find(parent,file),file,
                                                               message,data);
                      active=0;
                     }
                     break;


  case OM_FINDMASTER:
                     if((*((int*)data))==groupp->master)
                     {
                      *((int*)data)=groupp->key;
                      *message=0;
                     }
                     break;



   case   OM_SNAPBOX:
                     snapbox=((snapfn)data);
                     snapbox(groupp->key,&groupp->transform,
                                         &groupp->fcurve.box);
                     break;

 }

 if(!err) err=objectchainmessage(child,file,message,data);

 return(err);
}



static os_error * groupdb(groupstr * groupp,filestr * file,
                                             viewstr* view,objectdbstr * cont)
{
 switch(cont->type)
 {

  case ODB_OBJECTNAME:
                      strcpy(cont->name,"Group");
                      break;

  case ODB_OBJECTNAMEPLURAL:
                      strcpy(cont->name,"groups");
                      break;


    case ODB_MODIFY:
                    opengroupbox((sobjectstr*)groupp,file,view);
                    break;
 }
 return(NULL);
}


/*****************************************************************************/
/* script functions */

int  canungroup(int fp)
{
 int         object;
 groupstr  * groupp;

 if(nselected(&object,menufile)==1)
 {
  groupp=(groupstr*)fob_find(object,menufile);
  return(groupp->type==OGROUP && 
         !groupp->master      && 
         !embededobject(object,menufile));
 }
 else return(0);

 USE(fp);
}


int cangroup(int fp)
{
 int            i;
 int            master;
 objectbitstr   bits;

 if(nselected(NULL,menufile)>1)
 {
  for(i=0;i<selectn;i++)
  {
   objectmaster(selection[i].object,menufile,&master,&bits);

/*   dprintf(1,"nogr=%d master=%d embed=%d",bits.nogroup,master,
                  embededobject(selection[i].object,menufile)); */

   if(bits.nogroup) break;
   if(master) break;
   if(embededobject(selection[i].object,menufile)) break;
  }

  return(i>=selectn);
 }

 return(0);

 USE(fp);
}


/* tagged to UGRESTORE */

static os_error * restoreobject(filestr * file)
{
 os_error * err;
 int        object;
 int        after;
 int        parent;

 err=undoread(file,&object,sizeof(int));
 undoread(file,&parent,sizeof(int));
 undoread(file,&after,sizeof(int));

 insafter(object,after,parent,0,file);

 undowritefn(file,UGDELINK,&object,sizeof(int));

 return(err);
}


/* tagged to UGDELINK */

static os_error * delinkobject(filestr * file)
{
 os_error   * err;
 int          object;
 int          after;
 int          parent;
 sobjectstr * sob;

 err=undoread(file,&object,sizeof(int));

 sob=(sobjectstr*)fob_find(object,file);
 after=sob->chain.prev;
 parent=sob->chain.parent;

 undowritefn(file,UGRESTORE,&object,sizeof(int));
 undowrite(file,&parent,sizeof(int));
 undowrite(file,&after,sizeof(int));

 unlinkobject(object,0,file);

 return(err);
}


static os_error * savedelink(filestr * file,int object)
{
 os_error   * err;
 sobjectstr * sob;
 int          after;
 int          parent;

 err=NULL;

 sob=(sobjectstr*)fob_find(object,file);

 after=sob->chain.prev;
 parent=sob->chain.parent;

 undowritefn(menufile,UGRESTORE,&object,sizeof(int));
 undowrite(menufile,&parent,sizeof(int));
 undowrite(menufile,&after,sizeof(int));

 unlinkobject(object,0,file);

 return(err);
}



void groupselection(int fp)
{
 os_error   * err;
 int          i;
 int          object;
 int          after;
 int          parent;
 int          replace;
 int          rafter;
 int          rparent;
 curvestr     c;
 sobjectstr * sob;
 objectloc    loc;
 int        * list;
 int          listn;
 int          page;
 int          section;

 err=NULL;

/* dprintf(0,"gr1"); */

 if(cangroup(fp))
 {
  makeaselectionlist(&list,&listn,&section,&page);

  for(i=0;i<listn;i++)
  {
   if(embededobject(list[i],menufile))
   {
    err=geterror(EOB00);
    break;
   }
  }

 /* dprintf(0,"gr2"); */

  if(!err) err=modfile(menufile,section);
  if(!err)
  {
   undostart(menufile);

   clearselectionlo();

   /* by some means determine which object the group will replace */

   sortselectionlist(menufile,&list,listn);

   /* replace is now the last member of slist */
   /* scan back till you find something in slist whose predecessor is not */
   /* it is vital that rafter/rparent is not in the group of things */

   replace=list[listn-1];


  /* dprintf(0,"gr3 %d",replace); */

   while(replace)
   {
    sob=(sobjectstr*)fob_find(replace,menufile);
    object=sob->chain.prev;
    if(!inslist(object,list,listn)) break;
    replace=object;
   }

   sob=(sobjectstr*)fob_find(replace,menufile);
   rafter=sob->chain.prev;
   rparent=sob->chain.parent;

   memset(&c,0,sizeof(curvestr));
   c.type=CBOX;

 /* getslistbox(&c.box,menufile); */

   for(i=0;i<listn;i++)
   {
    object=list[i];
    savedelink(menufile,object);
   }

   loc.file=menufile;
   loc.chain=0;
   loc.parent=rparent;
   loc.after=rafter;
   loc.page=page;
   loc.section=section;

   objectcreate(&loc,OGROUP,&c);

   parent=loc.object;
   after=0;
  
   for(i=0;i<listn;i++)
   {
    object=list[i];
    insafter(object,after,parent,0,menufile);
    after=object;
    parent=0;
    undowritefn(loc.file,UGDELINK,&object,sizeof(int));
   }

   groupsize(loc.object,menufile);

   selectobject(loc.object,loc.page,loc.section,menufile);

   undoend(menufile);
  }
  clearaselectionlist(&list);
 }

 report(err);
}



void ungroupselection(int fp)
{
 os_error  * err;
 int         group;
 groupstr  * groupp;
 int         after;
 int         parent;
 int         object;
 int         section;
 int         page;

 err=NULL;

 if(canungroup(fp))
 {
  section=selsectk;
  page=selpagek;

  err=modfile(menufile,section);
  if(!err)
  {
   undostart(menufile);

   nselected(&group,menufile);

   clearselectionlo();

   groupp=(groupstr*)fob_find(group,menufile);

   after=groupp->chain.prev;
   parent=groupp->chain.parent;

   while(1)
   {
    groupp=(groupstr*)fob_find(group,menufile);
    object=groupp->chain.child;
    if(!object) break;

    savedelink(menufile,object);

    insafter(object,after,parent,0,menufile);
    parent=0;
    after=object;

    undowritefn(menufile,UGDELINK,&object,sizeof(int));

    selectobject(object,page,section,menufile);
   }

   objectdelete((objhdrstr*)fob_find(group,menufile),0,menufile);

   undoend(menufile);
  }
 }

 zreport(err);
}

/*****************************************************************************/

int  cangoforward(int fp)
{
 int          object;
 sobjectstr * sob;

 if(nselected(&object,menufile)==1)
 {
  sob=(sobjectstr*)fob_find(object,menufile);
  return(sob->chain.next);
 }
 else return(0);

 USE(fp);
}


/* tagged to UFORWARD */

static os_error * undoobjectforward(filestr * file)
{
 os_error      * err;
 sobjectstr    * sob;
 int             object;
 int             oldafter;
 int             oldparent;
 int             after=0;
 int             parent=0;
 pagesectionstr  pagesection;

 event(EVENT_SHIFTSELECTION0,file,0);

 err=undoread(file,&object,sizeof(int));

 vxstart(file);
 frflstart(file);

 sob=(sobjectstr*)fob_find(object,file);

 frameflow(&sob,file);
 frfladd(&sob,file);
 vxadd(&sob,file);

 oldafter=sob->chain.prev;
 oldparent=sob->chain.parent;

          err=undowritefn(file,UBACK,&object,sizeof(int));
 if(!err) err=undowrite(file,&oldafter,sizeof(int));
 if(!err) err=undowrite(file,&oldparent,sizeof(int));
 if(!err) err=undoread(file,&after,sizeof(int));
 if(!err) err=undoread(file,&parent,sizeof(int));
 if(!err) err=undoread(file,&pagesection,sizeof(pagesectionstr));
 if(!err) err=undowrite(file,&pagesection,sizeof(pagesectionstr));

 unlinkobject(object,0,file);
 insafter(object,after,parent,0,file);

 frflflush(file);
 vxflush(file,VRBACK,pagesection.page,pagesection.section);

 frflvalid(file);

 frflscrap(file);
 vxscrap();

 event(EVENT_SHIFTSELECTION1,file,0);

 return(err);
}


static os_error * objectforwards(int n)
{
 os_error     * err;
 int            object;
 sobjectstr   * sob;
 int            oldafter;
 int            oldparent;
 int            after;
 int            parent;
 pagesectionstr pagesection;


 err=NULL;

 if(cangoforward(0))
 {
  err=modfile(menufile,selsectk);
  if(!err)
  {
   undostart(menufile);

   event(EVENT_SHIFTSELECTION0,menufile,0);

   vxstart(menufile);
   frflstart(menufile);

   nselected(&object,menufile);
   sob=(sobjectstr*)fob_find(object,menufile);


   oldafter=sob->chain.prev;
   oldparent=sob->chain.parent;

            err=undowritefn(menufile,UFORWARD,&object,sizeof(int));
   if(!err) err=undowrite(menufile,&oldafter,sizeof(int));
   if(!err) err=undowrite(menufile,&oldparent,sizeof(int));

   pagesection.page=selpagek;
   pagesection.section=selsectk;

   if(!err) err=undowrite(menufile,&pagesection,sizeof(pagesectionstr));

   if(n<0) after=endofchain(object,menufile);
   else
   {
    after=object;
    while(n--)
    {
     sob=(sobjectstr*)fob_find(after,menufile);
     if(sob->chain.next) after=sob->chain.next;
     else                break;
    }
   }

   parent=0;

   unlinkobject(object,0,menufile);
   insafter(object,after,parent,0,menufile);

   sob=(sobjectstr*)fob_find(object,menufile);
   frameflow(&sob,menufile);
   frfladd(&sob,menufile);
   vxadd(&sob,menufile);

   frflflush(menufile);
   vxflush(menufile,VRBACK,selpagek,selsectk);

   frflvalid(menufile);

   frflscrap(menufile);
   vxscrap();

   event(EVENT_SHIFTSELECTION1,menufile,0);

   undoend(menufile);
  }
 }

 return(err);
}



void objectforward(int fp)
{
 zreport(objectforwards(-1));
 USE(fp);
}

void objectforwardn(int fp)
{
 zreport(objectforwards(stack[fp]));
}


/****************************************************************************/

int  cangobackward(int fp)
{
 int          object;
 sobjectstr * sob;

 if(nselected(&object,menufile)==1)
 {
  sob=(sobjectstr*)fob_find(object,menufile);
  return(sob->chain.prev);
 }
 else return(0);

 USE(fp);
}


/* tagged to UBACK */

static os_error * undoobjectbackward(filestr * file)
{
 os_error        * err;
 sobjectstr      * sob;
 int               object;
 int               oldafter;
 int               oldparent;
 int               after=0;
 int               parent=0;
 pagesectionstr    pagesection;


 event(EVENT_SHIFTSELECTION0,file,0);

 err=undoread(file,&object,sizeof(int));

 vxstart(file);
 frflstart(file);

 sob=(sobjectstr*)fob_find(object,file);

 oldafter=sob->chain.prev;
 oldparent=sob->chain.parent;

           err=undowritefn(file,UFORWARD,&object,sizeof(int));
  if(!err) err=undowrite(file,&oldafter,sizeof(int));
  if(!err) err=undowrite(file,&oldparent,sizeof(int));
  if(!err) err=undoread(file,&after,sizeof(int));
  if(!err) err=undoread(file,&parent,sizeof(int));
  if(!err) err=undoread(file,&pagesection,sizeof(pagesectionstr));
  if(!err) err=undowrite(file,&pagesection,sizeof(pagesectionstr));

 unlinkobject(object,0,file);
 insafter(object,after,parent,0,file);

 sob=(sobjectstr*)fob_find(object,file);
 frameflow(&sob,file);
 frfladd(&sob,file);
 vxadd(&sob,file);

 frflflush(file);
 vxflush(file,VRBACK,pagesection.page,pagesection.section);

 frflvalid(file);

 frflscrap(file);
 vxscrap();

 event(EVENT_SHIFTSELECTION1,file,0);

 return(err);
}




/* functions should use object move */


static os_error * objectbackwards(int n)
{
 os_error       * err;
 int              object;
 sobjectstr     * sob;
 int              oldafter;
 int              oldparent;
 int              after;
 int              parent;
 pagesectionstr   pagesection;

 err=NULL;

 if(cangobackward(0)) 
 {
  err=modfile(menufile,selsectk);
  if(!err)
  {
   undostart(menufile);

   event(EVENT_SHIFTSELECTION0,menufile,0);

   vxstart(menufile);

   frflstart(menufile);

   nselected(&object,menufile);
   sob=(sobjectstr*)fob_find(object,menufile);
   frameflow(&sob,menufile);
   frfladd(&sob,menufile);
   vxadd(&sob,menufile);

   oldafter=sob->chain.prev;
   oldparent=sob->chain.parent;

            err=undowritefn(menufile,UBACK,&object,sizeof(int));
   if(!err) err=undowrite(menufile,&oldafter,sizeof(int));
   if(!err) err=undowrite(menufile,&oldparent,sizeof(int));

   pagesection.page=selpagek;
   pagesection.section=selsectk;

   if(!err) err=undowrite(menufile,&pagesection,sizeof(pagesectionstr));

   if(n<0) after=startofchain(object,menufile);
   else
   {
    after=object;
    while(n--)
    {
     sob=(sobjectstr*)fob_find(after,menufile);
     if(sob->chain.prev) after=sob->chain.prev;
     else                break;
    }
   }

   sob=(sobjectstr*)fob_find(after,menufile);
   after=sob->chain.prev;
   parent=sob->chain.parent;

   unlinkobject(object,0,menufile);
   insafter(object,after,parent,0,menufile);

   frflflush(menufile);
   vxflush(menufile,VRBACK,selpagek,selsectk);

   frflvalid(menufile);
  
   frflscrap(menufile);
   vxscrap();

   event(EVENT_SHIFTSELECTION1,menufile,0);

   undoend(menufile);
  }
 }
 return(err);
}


void objectbackward(int fp)
{
 zreport(objectbackwards(-1));
 USE(fp);
}


void objectbackwardn(int fp)
{
 zreport(objectbackwards(stack[fp]));
}


/*****************************************************************************/

static textflowstr defflow=
{
 0,
 FRFBOTH,
 FRONONE,
};


static os_error * groupflow(groupstr * groupp,filestr * file,
                                             objectflowstr * flow,int read)
{
 os_error * err;

 err=NULL;

 if(read)
 {
  flow->flow=defflow;
  flow->tr=groupp->transform;
 }
 else
 {


 }

 return(err);

 USE(file);
 USE(groupp);
}




static os_error * groupsetflow(groupstr * groupp,sobjectstr * fix,
                                      filestr * file,int * end)
{
 os_error   * err;
 sobjectstr * childp;
 int          child;

 err=NULL;


 if(fix->key==groupp->key)
 {
  *end=1;
 }
 else
 {

  /* handle consequences for this object, then apply to children */

  child=groupp->chain.child;
  while(child)
  {
   childp=(sobjectstr*)fob_find(child,file);
   if(childp)
   {
    child=childp->chain.next;

    err=objectsetflow((objhdrstr*)childp,(objhdrstr*)fix,file,end);
    if(err) break;
    if(*end) break;
   }
   else break;
  }
 }

 return(err);
}



static os_error * masterobject(groupstr * groupp,filestr * file,int * master,
                                                          objectbitstr * bits)
{
 *master=groupp->master;
 memset(bits,0,sizeof(objectbitstr));
 bits->local=groupp->flags.local;
 bits->locked=groupp->flags.locked;
 bits->canmakelocal=1;
 return(NULL);
 USE(file);
}

/*****************************************************************************/
/* this is the trivial operation */

/* tagged to ULMXLOCAL */

static os_error *  unlmakelocal(filestr * file)
{
 os_error   * err;
 int          object;
 groupstr   * groupp=NULL;

          err=undoread(file,&object,sizeof(int));
 if(!err) err=undowritefn(file,ULMXMASTER,&object,sizeof(int));
 if(!err) err=fob_mod(object,file,(fobjptr)&groupp);
 if(!err)
 {
  groupp->flags.local=1;
 }

 return(err);
}


/* tagged to ULMXMASTER */

static os_error *  unlmakemaster(filestr * file)
{
 os_error   * err;
 int          object;
 groupstr   * groupp=NULL;

          err=undoread(file,&object,sizeof(int));
 if(!err) err=undowritefn(file,ULMXLOCAL,&object,sizeof(int));
 if(!err) err=fob_mod(object,file,(fobjptr)&groupp);
 if(!err)
 {
  groupp->flags.local=0;
 }

 return(err);
}


static os_error * makeobjectlocal(groupstr * groupp,filestr * file)
{
 os_error    * err;
 int           child;
 chainobject * childp;
 int           master;


 err=NULL;


 master=groupp->master;

 if(master)
 {
  if(!groupp->flags.local)  
  {
   groupp->flags.local=1;
   undowritefn(file,ULMXMASTER,&groupp->key,sizeof(int));

   child=groupp->chain.child;
   while(child)
   {
    err=objectlocal(child,file,MAKELOCAL);

    childp=(chainobject*)fob_find(child,file);
    if(childp) child=childp->chain.next;
    else       break;                                                              }
  }
 }

 return(err);
}



/*****************************************************************************/

typedef struct gmdatastr
{
 int          group;
 transformstr tr;

} gmdatastr;


static gmdatastr xgd;


/* tagged to UGMASTERS */
/* this function is always called last in undo/redo */

static os_error * ugmasters(filestr * file)
{
 groupstr * groupp;
 gmdatastr  ogd;



 fob_mod(xgd.group,file,(fobjptr)&groupp);

 ogd.group=xgd.group;
 ogd.tr=groupp->transform;

 groupp->flags.local^=1;
 groupp->transform=xgd.tr;

/* dprintf(5,"local=%d",groupp->flags.local); */


 groupsize(xgd.group,file);
 return(undowritefn(file,UGMASTERF,&ogd,sizeof(gmdatastr)));
}


/* tagged to UGMASTERF */
/* this function is always called first in undo/redo */

static os_error * ugmasterf(filestr * file)
{
 undoread(file,&xgd,sizeof(gmdatastr));
 return(undowritefn(file,UGMASTERS,NULL,0));
}


static os_error * makeobjectmaster(groupstr * groupp,filestr * file)
{
 os_error    * err;
 int           child;
 int           master;
 chainobject * childp;

 err=NULL;

 xgd.group=groupp->key;
 child=groupp->chain.child;
 master=groupp->master;

 groupp=(groupstr*)fob_find(master,file);
 xgd.tr=groupp->transform;

 undowritefn(file,UGMASTERS,NULL,0);

 while(child)
 {
  err=objectlocal(child,file,MAKEMASTER);

  childp=(chainobject*)fob_find(child,file);
  if(childp) child=childp->chain.next;
  else       break;
 }

 ugmasters(file);

 return(err);
}



static os_error * forceobjectmaster(int mode,groupstr * groupp,filestr * file)
{
 os_error    * err;
 int           child;
 chainobject * childp;
 int           group;

 err=NULL;

 group=groupp->key;

 child=groupp->chain.child;
 while(child)
 {
  err=objectlocal(child,file,mode);

  childp=(chainobject*)fob_find(child,file);
  if(childp) child=childp->chain.next;
  else       break;
 }

 groupsize(group,file);

 return(err);
}




static os_error * uudm(filestr * file)
{
 mrefstr      mref;
 objectloc    src;
 objectloc    dest;
 transformstr tr;

 undoread(file,&mref,sizeof(mrefstr));
 mpageundelete(&mref,0,&src,&dest,&tr,file);
 groupcopyc(&src,&dest,&tr);
 return(undowritefn(file,URDM,&dest.object,sizeof(int)));
}


static os_error * deletemaster(groupstr * groupp,filestr * file)
{
 int     group;
 mrefstr mref;

 group=groupp->key;

 mref.master=groupp->master;
 mref.after=groupp->chain.prev;

 groupscrap(group,file,DELETESCRAP|DELETEUNLINK);

 return(undowritefn(file,UUDM,&mref,sizeof(mrefstr)));
}


static os_error * urdm(filestr * file)
{
 int        group;
 groupstr * groupp;

 undoread(file,&group,sizeof(int));
 groupp=(groupstr*)fob_find(group,file);
 return(deletemaster(groupp,file));
}






static os_error * localobject(groupstr * groupp,filestr * file,int makelocal)
{

/* dprintf(6,"makelocal=%d",makelocal); */

 if(makelocal==MAKELOCAL)    return(makeobjectlocal(groupp,file));
 else
 if(makelocal==MAKEMASTER)   return(makeobjectmaster(groupp,file));
 else
 if(makelocal==DELETEMASTER) return(deletemaster(groupp,file));
 else                        return(forceobjectmaster(makelocal,groupp,file));
}


/*****************************************************************************/




static os_error * shapeobject(groupstr * groupp,int shape,
                                              int data,filestr * file)
{
 os_error    * err;
 int           child;
 chainobject * childp;

 err=NULL;

 child=groupp->chain.child;
 while(child)
 {
  err=objectshape(child,shape,data,file);
  if(err) break;
  childp=(chainobject*)fob_find(child,file);
  child=childp->chain.next;
 }

 return(err);
}



/*****************************************************************************/



static colourvalue gcv={OURTRANS};


static os_error * oreaddata(int action,groupstr * groupp,
                                             filestr * file,void * data)

{

 switch(action)
 {
  case SELECTINFO:
                  ((outlineinfostr*)data)->fcurve=groupp->fcurve;
                  ((outlineinfostr*)data)->tr=groupp->transform;
                  ((outlineinfostr*)data)->frame=0;
                  break;

  case  OREADFILL:
                  *((colourvalue*)data)=gcv;
                  break;
 }

 return(NULL);

 USE(file);
}







/*****************************************************************************/

static objectstr groupobject=
{
 OGROUP,                            /* type        */
 NULL,                              /* new         */
 (objectcopyfn)groupcopy,           /* copy        */
 (objectredrawfn)groupredraw,       /* redraw      */
 (objectcreatefn)groupcreate,       /* create      */
 (objectfindfn)groupfind,           /* find        */
 (objectselectfn)groupselect,       /* select      */
 (objecttransformfn)grouptransform, /* transform   */
 NULL,                              /* load type   */
 NULL,                              /* load        */
 NULL,                              /* load post   */
 (objectmessagefn)groupmessage,     /* message     */
 (objectdbfn)groupdb,               /* object db's */
 (objectflowfn)groupflow,           /* text flow   */
 (objectdeletefn)groupdelete,       /* delete */
 NULL,                              /* text reflow */
 (objectsetflowfn)groupsetflow,     /* object set flow */
 (objectmasterfn)masterobject,      /* master object?  */
 (objectlocalfn)localobject,        /* local object    */
 (objectshapefn)shapeobject,        /* change shape    */
 (objectreadfn)oreaddata,           /* read data */
 NULL,                              /* x12 */

};


#define UNDOS 15

static undofnstr undos[UNDOS]=
{
 ungrouptransform,   NULL,    UGTRAN,     "{UTRANS}",
 undoobjectbackward, NULL,    UBACK,      "{UFRONT}",
 undoobjectforward,  NULL,    UFORWARD,   "{UBACK}",
 ugrecreate,         ugtrash, UGRECREATE, NULL,
 ugdelete,           NULL,    UGDELETE,   NULL,
 restoreobject,      NULL,    UGRESTORE,  NULL,
 delinkobject,       NULL,    UGDELINK,   NULL,
 unlmakelocal,       NULL,    ULMXLOCAL,  NULL,
 unlmakemaster,      NULL,    ULMXMASTER, NULL,
 ugmasters,          NULL,    UGMASTERS,  NULL,
 ugmasterf,          NULL,    UGMASTERF,  NULL,
 ungmove,            NULL,    UGMOVE,     NULL,
 uudm,               NULL,    UUDM,       NULL,
 urdm,               NULL,    URDM,       NULL,
 ungroupupdate,      NULL,    UGUPDATE,   NULL,
};




os_error * groupinit(void)
{
 os_error * err;
 err=undoaddblock(undos,UNDOS);
 if(!err) err=objectregister(&groupobject);
 dsaddblock(dsdefs,sizeof(dsdefs)/sizeof(dscreator));
 return(err);
}


