/*->h.txstr */



/* largest style chain - used to allocate static memory */

#define MAXCHAIN 80


/* types of vertical alignment */

#define VTOP     0
#define VBOTTOM  1
#define VCENTRE  2
#define VJUSTIFY 3


/* types of first line */

#define FBLACCENT   0
#define FBLASCENDER 1
#define FBLFIT      2
#define FBLGRID     3


/* scopes for user styles */

#define SCOPECURR  0
#define SCOPEPARA  1
#define SCOPEWORD  2
#define SCOPELINE  3
#define SCOPESTORY 4

#define SCOPECURRK 100




#define STYLENAMELEN 20


#define SUPS 1
#define SUBS 2


#define TITLE     1
#define ALLCAPS   2
#define ALLLC     3


#define LEFT    0
#define CENTRE  1
#define RIGHT   2
#define JUSTIFY 3


/* NB these bits are needed to get redraws right if styles have gone */

#define ISLEFT  0x0     /* line is left  justified          */
#define ISCENT  0x1     /* line is centered                 */
#define ISRIGHT 0x2     /* line is right justified          */
#define ISJUST  0x3     /* line is justified                */


#define ISSPACE 0x0     /* line ends with a space           */
#define ISBREAK 0x1     /* line ends with a break           */
#define ISNL    0x2     /* line ends with a NL              */
#define ISTEXT  0x3     /* line ends with ordinary char     */
#define ISHYP   0x4     /* hyphenated                       */

#define ISDROP  0x5     /* line is a drop cap               */


#define MINDROP 0x2
#define MAXDROP 0x1F


#define MINKEEP 0x0
#define MAXKEEP 0x1F


#define MINUSINT 0x10






#define JUSTWORDSCALE 0x5
#define JUSTWORDMIN   25
#define JUSTWORDMAX   1000
#define JUSTLETTMIN   -50
#define JUSTLETTMAX   100


#define MAXBASESHIFT  (10*0x10000)
#define MINBASESHIFT  (-10*0x10000)

#define MAXTRACK      1000
#define MINTRACK     -1000

#define MAXHSCALE     (10*0x10000)
#define MINHSCALE     (0x10000/10)

#define MINPTSIZE     1
#define MAXPTSIZE     1000



/* types of leading */

/* absolute */

#define LEADABS  0x1

/* percent */

#define LEADPC   0x2

/* incremental */

#define LEADDL   0x3


#define BODYTEXT 1

#define BULLETNOSTYLE   0xFFFF
#define BULLETSELFSTYLE 0


/* types of indent */

#define INDENTABS    0
#define INDENTDELTA  1
#define INDENTOFFSET 2



#define MAXBLEVEL 8

// document compatability level needed for Unicode

#define UNICOMPAT 5


/*****************************************************************************/


#define MAXCOLUMNS 10


typedef struct columnstr
{
 unsigned int mode:2;
 unsigned int guides:1;
 unsigned int guidewidth:24;
 int          width;
 colourvalue colour;

} columnstr;



typedef struct flowchainstr
{
 int next;
 int prev;
} flowchainstr;



typedef struct textvbox
{
 int            key;
 int            type;

 int            down;

 int            nc;     /* number of curves */

 curvestr       curve[1];

} textvbox;



typedef struct texsub
{
 int            bytes;     /* bytes in this box            */
 int            vlines;
 int            vbox;      /* points to text vbox's        */
 int            vdelta;    /* space for vertical alignment */
} texsub;



typedef struct texhdr
{
 unsigned int header:30;
 unsigned int headerdata:1;
 unsigned int headerfull:1;

 unsigned int footer:30;
 unsigned int footerdata:1;
 unsigned int footerfull:1;
} texhdr;




typedef struct texsrc
{
 int source;       /* where the text is stored         */
 int vsource;      /* where vlines are stored          */
} texsrc; 



typedef struct textdatastr
{
 unsigned int   columns:5;
 unsigned int   header:1;
 unsigned int   i:25;      /* inset              */
 unsigned int   changei:1;

 unsigned int   valign:4;
 unsigned int   repeat:1;  /* repeating frame    */
 unsigned int   footer:1;
   signed int   radius:26; /* was unsigned */

 colourvalue    colour;    /* background colour  */

 columnstr      columndata;
 transformstr   transform;

 fillstr        fill;

} textdatastr;



typedef struct unpackedtextdatastr
{
 int           columns;
 int           header;
 int           inset;
 int           changei;

 int           valign;
 int           repeat;
 int           footer;
 int           radius;

 colourvalue   colour;

 columnstr     columndata;

 fillstr       fill;

 int           lock;
 int           spread;

} unpackedtextdatastr;



typedef struct textchangebits
{

 unsigned int           xcolumns:1;
 unsigned int           xheader:1;
 unsigned int           xinset:1;
 unsigned int           xchangei:1;
 unsigned int           xvalign:1;
 unsigned int           xrepeat:1;
 unsigned int           xfooter:1;
 unsigned int           xradius:1;
 unsigned int           xcolour:1;
 unsigned int           xcolumndata:1;
 unsigned int           xfill:1;
 unsigned int           xlock:1;
 unsigned int           xspread:1;

 unsigned int           xtransform:1;

 unsigned int           xfirstline:1;

} textchangebits;




typedef struct textboxstr
{
 int            key;
 int            type;
 chainstr       chain;
 flagstr        flags;

 boxstr         box;
 transformstr   transform;

 curvestr       fcurve;    /* frame curve                      */
 curvestr       icurve;    /* internal curve                   */
 curvestr       ocurve;    /* external curve                   */
 textflowstr    textflow;
 int            master;    /* tied to something on master page */

 int            frame;     /* frame pointer                    */
 textdatastr    data;

 flowchainstr   fchain;    /* flow chain                       */

 /* variable size data                    */
 /*                       n=1 or 3 texsrc */
 /* followed by           1        texhdr */
 /* followed by           m        texsub */

} textboxstr;



/*****************************************************************************/


typedef struct actionbits
{
 unsigned int bold:1;
 unsigned int italic:1;
 unsigned int underline:1;
 unsigned int reverse:1;
 unsigned int supersub:1;
 unsigned int strikeout:1;
 unsigned int rulevert:1;
 unsigned int rulehoriz:1;
 unsigned int font:1;
 unsigned int fcolour:1;
 unsigned int bcolour:1;
 unsigned int size:1;
 unsigned int hscale:1;
 unsigned int track:1;
 unsigned int baseshift:1;
 unsigned int align:1;
 unsigned int before:1;
 unsigned int after:1;
 unsigned int leading:1;
 unsigned int left:1;
 unsigned int right:1;
 unsigned int first:1;
 unsigned int xcase:1;
 unsigned int drop:1;
 unsigned int bullet:1;
 unsigned int hyphenation:1;
 unsigned int akern:1;         /* auto kern */
 unsigned int wordwrap:1;      /* word wrap */
 unsigned int index:1;
 unsigned int contents:1;
 unsigned int layer:1;
 unsigned int baselinelock:1;  /* 32 bits */



 unsigned int smlcaps:1;
 unsigned int keeplines:1;
 unsigned int keeppara:1;
 unsigned int keepnext:1;
 unsigned int language:1;
 unsigned int tab:1;
 unsigned int rotate:1;
 unsigned int skew:1;
 unsigned int justification:1;
 unsigned int spare:23;        /* used to keep spare bits zeroed */

} actionbits;



// max action bit index + 1

#define MAXEFFECT 43



#define LEADERLEN 3

#define LEADERMAXLEN 16

#define DECTABMAXLEN 16

#define MAXHYPHENCHAR	12
#define MAXDECTABCHAR	12


typedef struct datawords
{
 unsigned int         font:16;
 unsigned int         mfont:16;   /* font map key for this font */

          colourvalue fcolour;
          colourvalue bcolour;
          colourvalue ulcolour;
          colourvalue stcolour;
          colourvalue vrulecolour;
          colourvalue hrulecolour;

 unsigned int         ulthickness:24;
 unsigned int         language:8;

 unsigned int         stthickness:24;
 unsigned int         dropno:8;

 short    int         rotate;        /* 0..32K -> 0..179 */
 short    int         skew;

 unsigned int         xpts:24;
 unsigned int         hpmin:8;

 unsigned int         ypts:24;
 unsigned int         hpbefore:8;

 signed   int         track:24;
 unsigned int         hpafter:8;

 signed   int         baseshift:24;
 unsigned int         bulletbefore1:8;

 unsigned int         ascent:24;    /* from font man - low bit lead */
 unsigned int         bulletafter1:8;

          char        leaderchar[3]; /* tab leader char - should be a char */
          char        bulletdata1;

 unsigned int         height:24;  /* a number which comes from font man. */
 unsigned int         bulletstart:8;

 unsigned int         vrulethickness:24;
 signed   int         justlettermin:8;

 unsigned int         hrulethickness:24;
 signed   int         justlettermax:8;

 signed   int         hpzone:24;
 unsigned int         hpconsec:8;

 unsigned int         justflushzone:24;
 unsigned int         justwordmax:8;

 signed   int         ruleaboveshift:24;
 unsigned int         justwordmin:8;

 signed   int         rulebelowshift:24;
 unsigned int         bold:1;
 unsigned int         italic:1; 
 unsigned int         underline:1;
 unsigned int         reverse:1;
 unsigned int         supersub:2;  /* 0 normal, 1 super, 2 sub  */
 unsigned int         strikeout:1;
 unsigned int         rulevert:1;

 signed   int         before:24;
 unsigned int         rulehoriz:1;
 unsigned int         align:3;     /* 0 left, 1 centre, 2 right, 3 justify */
 unsigned int         lefttype:1;  /* 0=from frame, 1=from obstruction */
 unsigned int         righttype:2; /* ... 2 from left */ 
 unsigned int         bullet:1;

 signed   int         after:24;
 unsigned int         xcase:3;
 unsigned int         drop:1;
 unsigned int         ultype:2;
 unsigned int         sttype:2;

 unsigned int         bulletstyle:16;
 unsigned int         layer:5;
 unsigned int         keeplinesno:5;
 unsigned int         hyphenation:1;
 unsigned int         akern:1;         /* auto kern */
 unsigned int         wordwrap:1;      /* word wrap */
 unsigned int         index:1;
 unsigned int         contents:1;
 unsigned int         baselinelock:1;

 signed   int         leading:24;         /* for percent, times 64K */
 unsigned int         smlcaps:1;
 unsigned int         keeplines:1;
 unsigned int         keeppara:1;
 unsigned int         keepnext:1;       /* 30 */
 unsigned int         uldouble:1;
 unsigned int         ulword:1;
 unsigned int         ruleabove:1;
 unsigned int         rulebelow:1;

 unsigned int         lead:24;    /* a number which comes from font man. */
 unsigned int         ruleleft:1;
 unsigned int         ruleright:1;
 unsigned int         hpbreaklast:1;
 unsigned int         hpbreakcaps:1;
 unsigned int         leadingtype:3;  /* absolute, percent, or incremental */
 unsigned int         bulletornumber:1;

 unsigned int         tab:18;          /* point to tabs */
 unsigned int         bullettab:1;
 unsigned int         justification:1;
 unsigned int         justsingle:1;
 unsigned int         aftertype:1;
 unsigned int         beforetype:1;
 unsigned int         spare:9;

          int         leftshift;

          int         rightshift;

          int         firstshift;

//        int         ruleleftshift;
 char                 bulletbefore2;
 char                 bulletafter2;
 char                 bulletdata2;
 char                 ruleleftshift;


        int         rulerightshift;

//	int                  extension;


} datawords;






typedef struct xstylestr
{
 int           n;                       /* unique number */
 actionbits    action;
 datawords     data;
} xstylestr;




typedef struct absstylestr
{
 int       chain;                /* point at chain of user styles */
 unsigned int       count:24;             /* no of uses of this abs style  */
 unsigned int       changed:1;
 xstylestr style;
} absstylestr;




typedef struct absstylebasestr
{
 int         key;
 int         type;

 int         spare[4];

 int         n;
 unsigned int         sequence:31;
 unsigned int         leaderfix:1;  /* calc ascender for old files */
 absstylestr absstyle[1];
} absstylebasestr;




typedef struct userstylestr
{
 char          name[STYLENAMELEN];
 signed int    key:16;                /* key press for this style */
 unsigned int           menu:1;                /* show on menu ? */
 unsigned int           scope:4;               /* paragraph style ? */
 unsigned int           nodelete:1;
 unsigned int           changed:1;
 unsigned int           sparebits0:9;
 unsigned int           rank:16;
 unsigned int           sparebits1:16;
 int           spare[3];
 xstylestr     style;

} userstylestr;




typedef struct userstylebasestr
{
 int           key;
 int           type;

 int           spare[4];

 int           n;
 unsigned int           sequence:31;
 unsigned int           leaderfix:1;    /* calc ascender for old files */
 userstylestr  userstyle[1];
} userstylebasestr;



/*****************************************************************************/

/* used to hold text */


typedef struct sourcestr
{
 int            key;
 int            type;
 int            len;
 int            maxlen;
 int            goffset;

 int            spare;

 char           data[1];
} sourcestr;


/* used to hold vlines */


typedef struct vsourcestr
{
 int            key;
 int            type;
 int            len;
 int            maxlen;
 int            goffset;
 int            fldstack;

 int            spare;

 char           data[1];
} vsourcestr;


#define MAXW(vline) (((vline)->line.w1)|(((vline)->line.w2)<<8)|(((vline)->line.w3)<<16))
#define MAXWX(vline) (((vline).line.w1)|(((vline).line.w2)<<8)|(((vline).line.w3)<<16))
#define SETMAXW(vline,w) (vline->line.w1=(w),vline->line.w2=(w)>>8,vline->line.w3=(w)>>16)

#define LY(yy) ((signed int)(((yy)>0x7C0000)?((yy)|0xFF800000):(yy)))


typedef union vlstr
{
 struct
 {
 unsigned  int     x:23;
 unsigned  int     w1:8;
 unsigned  int     xl:1;       /* spare 1 */

 unsigned  int     yy:23;
 unsigned  int     w2:8;
 unsigned  int     spare1:1;   /* spare 1 */


 unsigned  int     h:23;
 unsigned  int     w3:8;
 unsigned  int     spare2:1;   /* spare 1 */

 unsigned  int     w:23;
 unsigned  int     just:2;     /* justification bits                              */
 unsigned  int     strip:1;    /* flag if this vline is last  in horizontal strip */
 unsigned  int     font:1;     /* flag if this vline changes font                 */
 unsigned  int     eol:3;      /* line end                                        */
                      /* ends with hyphen?                               */
 unsigned  int     object:1;   /* flag if this vline contains object              */
 unsigned  int     bullet:1;   /* flag if this vline is bulleted                  */

 unsigned  int     style:18;
 unsigned  int     len:14;
 } line;


 struct
 {
 unsigned  int bits1:24;
 unsigned  int type:7;
 unsigned  int xl:1;
  int spare[4];

 } ext;


 struct
 {
  unsigned  int                tab:24;
  unsigned  int                type:7;
  unsigned  int                  xl:1;
  unsigned short int levels[MAXBLEVEL];

 } bseq;


} vlstr;



#define VLBULLET 0x1






/*****************************************************************************/


typedef struct stylechainstr
{
 int key;
 int offset;
 unsigned int count:24;
 unsigned int changed:1;
 int n;

} stylechainstr;



typedef struct cstylebasestr
{
 int           key;
 int           type;

 int           spare[4];

 int           n;
 int           sequence;

 stylechainstr stylechain[1];

} cstylebasestr;




typedef struct tabhdrstr   /* one of these for each tab ruler */
{
 int key;
 int offset;
 int count;
 int n;

} tabhdrstr;


typedef struct tabbasestr
{
 int         key;
 int         type;

 int         spare[4];

 int         n;              /* how many tab rulers */
 int         sequence;

 tabhdrstr   tabhdr[1];

} tabbasestr;


/*****************************************************************************/

typedef struct textsrcstr
{
 int          source;
 sourcestr  * sourcep;             /* source               */
 int          offset;              /* offset in source     */
 int          loffset;             /* start of line offset */

 int          vsource;
 vsourcestr * vsourcep;
 int          voffset;

 int          text;
 int          s;                   /* sub text box number              */
 int          u;                   /* unit 0==main 1==header 2==footer */
 int          usize;               /* size of unit data                */
 int          columns;             /* effective columns                */

 textboxstr * textp;

} textsrcstr;



typedef struct vector             /* position vector  */
{
 filestr    * file;

 int          section;            /* section          */
 int          page;               /* page             */

 textsrcstr   textsrc;

 int          boffset;
 int          bvoffset;

 int          astyle;             /* absolute style at start of line */

 int          x;                  /* coords of point */
 int          y;

 int          xxxxxx;

 int          valid;

} vector;



#define VALIDOFFSET 0x1
#define VALIDPOSN   0x2
#define VVECTOR    (VALIDOFFSET|VALIDPOSN)



/****************************************************************************/

typedef struct flowdatablock
{
 unsigned int type:31;                      /* type of flow, quick, delta, dirty */
 unsigned int bulletstate:1;
 int offset;                       
 int length;                       /* length                            */
 int delta;                        /* current delta value               */
} flowdatablock;


#define FLBLOCK 8


typedef struct flowdatastr
{
 int           key;
 int           type;
 vector        fv;                 /* vector for lowest point in flow */
 int           next;               /* link all flow stacks for file   */
 int           prev;
 int           sp;                 /* stack pointer */
 flowdatablock data[FLBLOCK];
} flowdatastr;



/**************************************************************************/

typedef struct txmodstr
{
 int offset;
 int size;

} txmodstr;


/**************************************************************************/



typedef struct bulletdatastr
{
 int tabshift;

 int levels[MAXBLEVEL];

} bulletdatastr;





/**************************************************************************/

/* EOB format */

/*  0 code EOB                                    */

/*  1 object reference                            */
/*  2                                             */
/*  3                                             */
/*  4                                             */

/*  5 X shift from start of line                  */
/*  6                                             */
/*  7                                             */

/*  8 Y shift of top left corner from top of line */
/*  9                                             */
/* 10                                             */

/* 11 Angle of the frame last embedded in         */
/* 12                                             */
/* 13                                             */
/* 14                                             */

/* 15 EOB height                                  */
/* 16                                             */
/* 17                                             */

/* 18 EOB width                                   */
/* 19                                             */
/* 20                                             */

/* 21 base shift %                                */
/* 22                                             */
/* 23                                             */
/* 24                                             */

/* 25 spare                                       */
/* 26                                             */
/* 27                                             */
/* 28                                             */
/* 29                                             */
/* 30                                             */
/* 31                                             */

